from collections.abc import Callable

import keras
import numpy as np

from bayesflow.types import Tensor

from .vjp import vjp


def jacobian(f: Callable[[Tensor], Tensor], x: Tensor, return_output: bool = False):
    """
    Compute the Jacobian matrix of f with respect to x.

    Parameters
    ----------
    f : callable
        The function to be differentiated.
    x : Tensor of shape (..., D_in)
        The input tensor to f.
    return_output : bool, optional
        Whether to return the output of f(x) along with the Jacobian matrix.
        Default: False

    Returns
    -------
    Tensor of shape (..., D_out, D_in)
        The Jacobian matrix of f with respect to x.

    2-tuple of tensors
        1. The output of f(x) (if return_output is True)
        2. Tensor of shape (..., D_out, D_in)
            The Jacobian matrix of f with respect to x.

    """
    fx, vjp_fn = vjp(f, x, return_output=True)

    batch_shape = keras.ops.shape(x)[:-1]
    batch_size = np.prod(batch_shape)

    rows = keras.ops.shape(fx)[-1]
    cols = keras.ops.shape(x)[-1]

    jac = keras.ops.zeros((*batch_shape, rows, cols))

    for col in range(cols):
        projector = np.zeros(keras.ops.shape(x), dtype=keras.ops.dtype(x))
        projector[..., col] = 1.0
        projector = keras.ops.convert_to_tensor(projector)

        # jac[..., col] = vjp_fn(projector)
        indices = np.stack(list(np.ndindex(batch_shape + (rows,))))
        indices = np.concatenate([indices, np.full((batch_size * rows, 1), col)], axis=1)
        indices = keras.ops.convert_to_tensor(indices)

        updates = vjp_fn(projector)
        updates = keras.ops.reshape(updates, (-1,))
        jac = keras.ops.scatter_update(jac, indices, updates)

    if return_output:
        return fx, jac

    return jac



import numpy as np
from scipy.integrate import odeint

from .benchmark_simulator import BenchmarkSimulator


class LotkaVolterra(BenchmarkSimulator):
    def __init__(
        self,
        X0: int = 30,
        Y0: int = 1,
        T: int | None = 20,
        subsample: int = 10,
        flatten: bool = True,
        obs_noise: float = 0.1,
        dt: float = None,
        rng: np.random.Generator = None,
    ):
        """Lotka Volterra simulated benchmark.
        See: https://arxiv.org/pdf/2101.04653.pdf, Task T.10

        Parameters
        ----------
        X0: int, optional, default: 30
            Initial number of prey species.
        Y0: int, optional, default: 1
            Initial number of predator species.
        T: int, optional, default: 20
            The duration (time horizon) of the simulation.
        subsample: int or None, optional, default: 10
            The number of evenly spaced time points to return.
            If None, no subsampling will be performed and all T timepoints will be returned.
        flatten: bool, optional, default: True
            A flag to indicate whether a 1D (`flatten=True`) or 2D (`flatten=False`)
            representation of the simulated data is returned.
        obs_noise: float, optional, default: 0.1
            The standard deviation of the log-normal likelihood.
        rng: np.random.Generator or None, optional, default: None
            An optional random number generator to use.
        """

        self.X0 = X0
        self.Y0 = Y0
        self.T = T
        if dt is None:
            dt = 1 / T
        self.dt = dt
        self.subsample = subsample
        self.flatten = flatten
        self.obs_noise = obs_noise
        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()

    def _deriv(self, x, t, alpha, beta, gamma, delta):
        """Helper function for scipy.integrate.odeint."""

        X, Y = x
        dX = alpha * X - beta * X * Y
        dY = -gamma * Y + delta * X * Y
        return dX, dY

    def prior(self):
        """Generates a random draw from a 4-dimensional (independent) lognormal prior
        which represents the four contact parameters of the Lotka-Volterra model.

        Returns
        -------
        params : np.ndarray of shape (4, )
            A single draw from the 4-dimensional prior.
        """

        params = self.rng.lognormal(mean=[-0.125, -3, -0.125, -3], sigma=0.5)
        return params

    def observation_model(self, params: np.ndarray) -> np.ndarray:
        """Runs a Lotka-Volterra simulation for T time steps and returns `subsample` evenly spaced
        points from the simulated trajectory, given contact parameters `params`.

        Parameters
        ----------
        params      : np.ndarray of shape (2,)
            The 2-dimensional vector of disease parameters.

        Returns
        -------
        x : np.ndarray of shape (subsample, 2) or (subsample*2,) if `subsample is not None`,
            otherwise shape (T, 2) or (T*2,) if `subsample is None`.
            The time series of simulated predator and pray populations
        """

        # Create vector (list) of initial conditions
        x0 = self.X0, self.Y0

        # Unpack parameter vector into scalars
        alpha, beta, gamma, delta = params

        # Prepate time vector between 0 and T of length T
        t_vec = np.linspace(0, self.T, int(1 / self.dt))

        # Integrate using scipy and retain only infected (2-nd dimension)
        pp = odeint(self._deriv, x0, t_vec, args=(alpha, beta, gamma, delta))

        # Subsample evenly the specified number of points, if specified
        if self.subsample is not None:
            pp = pp[:: (self.T // self.subsample)]

        # Ensure minimum count is 0, which will later pass by log(0 + 1)
        pp[pp < 0] = 0.0

        # Add noise, decide whether to flatten and return
        x = self.rng.lognormal(np.log1p(pp), sigma=self.obs_noise)
        if self.flatten:
            return x.flatten()
        return x



import keras
import numpy as np
import pytest
from bayesflow.utils.numpy_utils import softmax


@pytest.fixture()
def var_names():
    return [r"$\beta_0$", r"$\beta_1$", r"$\sigma$"]


@pytest.fixture()
def random_samples_a():
    return np.random.normal(loc=0, scale=1, size=(5000, 8))


@pytest.fixture()
def random_samples_b():
    return np.random.normal(loc=0, scale=3, size=(5000, 8))


@pytest.fixture()
def random_estimates():
    return {
        "beta": np.random.standard_normal(size=(32, 10, 2)),
        "sigma": np.random.standard_normal(size=(32, 10, 1)),
    }


@pytest.fixture()
def random_targets():
    return {
        "beta": np.random.standard_normal(size=(32, 2)),
        "sigma": np.random.standard_normal(size=(32, 1)),
        "y": np.random.standard_normal(size=(32, 3, 1)),
    }


@pytest.fixture()
def random_priors():
    return {
        "beta": np.random.standard_normal(size=(64, 2)),
        "sigma": np.random.standard_normal(size=(64, 1)),
        "y": np.random.standard_normal(size=(64, 3, 1)),
    }


@pytest.fixture()
def model_names():
    return [r"$\mathcal{M}_0$", r"$\mathcal{M}_1$", r"$\mathcal{M}_2$"]


@pytest.fixture()
def true_models():
    true_models = np.random.choice(3, 100)
    true_models = np.eye(3)[true_models].astype(np.int32)
    return true_models


@pytest.fixture()
def pred_models(true_models):
    pred_models = np.random.normal(loc=true_models)
    pred_models = softmax(pred_models, axis=-1)
    return pred_models


@pytest.fixture()
def history():
    h = keras.callbacks.History()

    step = np.linspace(0, 1, 10_000)
    train_loss = (1.0 - step) ** 2 + np.random.normal(loc=0, scale=0.02, size=step.shape)
    validation_loss = 0.1 + (0.75 - step) ** 2 + np.random.normal(loc=0, scale=0.02, size=step.shape)

    h.history = {
        "loss": train_loss.tolist(),
        "val_loss": validation_loss.tolist(),
    }

    return h



import numpy as np

from bayesflow.utils.serialization import serializable, serialize

from .elementwise_transform import ElementwiseTransform


@serializable
class ConvertDType(ElementwiseTransform):
    """
    Default transform used to convert all floats from float64 to float32 to be in line with keras framework.

    Parameters
    ----------
    from_dtype : str
        Original dtype
    to_dtype : str
        Target dtype
    """

    def __init__(self, from_dtype: str, to_dtype: str):
        super().__init__()

        self.from_dtype = from_dtype
        self.to_dtype = to_dtype

    def get_config(self) -> dict:
        config = {
            "from_dtype": self.from_dtype,
            "to_dtype": self.to_dtype,
        }
        return serialize(config)

    def forward(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return data.astype(self.to_dtype, copy=False)

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return data.astype(self.from_dtype, copy=False)



from collections.abc import Mapping

import numpy as np

import keras

from bayesflow.types import Tensor
from bayesflow.utils import filter_kwargs, split_arrays, squeeze_inner_estimates_dict, logging
from bayesflow.utils.serialization import serializable

from .continuous_approximator import ContinuousApproximator


@serializable
class PointApproximator(ContinuousApproximator):
    """
    A workflow for fast amortized point estimation of a conditional distribution.

    The distribution is approximated by point estimators, parameterized by a feed-forward
    :py:class:`~bayesflow.networks.PointInferenceNetwork`. Conditions can be compressed by an optional summary network
    (inheriting from :py:class:`~bayesflow.networks.SummaryNetwork`) or used directly as input to the inference network.
    """

    def estimate(
        self,
        conditions: Mapping[str, np.ndarray],
        split: bool = False,
        **kwargs,
    ) -> dict[str, dict[str, np.ndarray | dict[str, np.ndarray]]]:
        """
        Estimates point summaries of inference variables based on specified conditions.

        This method processes input conditions, computes estimates, applies necessary adapter transformations,
        and optionally splits the resulting arrays along the last axis.

        Parameters
        ----------
        conditions : Mapping[str, np.ndarray]
            A dictionary mapping variable names to arrays representing the conditions
            for the estimation process.
        split : bool, optional
            If True, the estimated arrays are split along the last axis, by default False.
        **kwargs
            Additional keyword arguments passed to underlying processing functions.

        Returns
        -------
        estimates : dict[str, dict[str, np.ndarray or dict[str, np.ndarray]]]
            The estimates of inference variables in a nested dictionary.

            1. Each first-level key is the name of an inference variable.
            2. Each second-level key is the name of a scoring rule.
            3. (If the scoring rule comprises multiple estimators, each third-level key is the name of an estimator.)

            Each estimator output (i.e., dictionary value that is not itself a dictionary) is an array
            of shape (num_datasets, point_estimate_size, variable_block_size).
        """

        conditions = self._prepare_conditions(conditions, **kwargs)
        estimates = self._estimate(**conditions, **kwargs)
        estimates = self._apply_inverse_adapter_to_estimates(estimates, **kwargs)
        # Optionally split the arrays along the last axis.
        if split:
            estimates = split_arrays(estimates, axis=-1)
        # Reorder the nested dictionary so that original variable names are at the top.
        estimates = PointApproximator._reorder_estimates(estimates)
        # Remove unnecessary nesting.
        estimates = PointApproximator._squeeze_estimates(estimates)

        return estimates

    def sample(
        self,
        *,
        num_samples: int,
        conditions: Mapping[str, np.ndarray],
        split: bool = False,
        **kwargs,
    ) -> dict[str, dict[str, np.ndarray]]:
        """
        Draws samples from a parametric distribution based on point estimates for given input conditions.

        These samples will generally not correspond to samples from the fully Bayesian posterior, since
        they will assume some parametric form (e.g., multivariate normal when using the MultivariateNormalScore).

        Parameters
        ----------
        num_samples : int
            The number of samples to generate.
        conditions : Mapping[str, np.ndarray]
            A dictionary mapping variable names to arrays representing the conditions
            for the sampling process.
        split : bool, optional
            If True, the sampled arrays are split along the last axis, by default False.
            Currently not supported for :py:class:`PointApproximator` .
        **kwargs
            Additional keyword arguments passed to underlying processing functions.

        Returns
        -------
        samples : dict[str, np.ndarray or dict[str, np.ndarray]]
            Samples for all inference variables and all parametric scoring rules in a nested dictionary.

            1. Each first-level key is the name of an inference variable.
            2. (If there are multiple parametric scores, each second-level key is the name of such a score.)

            Each output (i.e., dictionary value that is not itself a dictionary) is an array
            of shape (num_datasets, num_samples, variable_block_size).
        """
        conditions = self._prepare_conditions(conditions, **kwargs)
        samples = self._sample(num_samples, **conditions, **kwargs)
        samples = self._apply_inverse_adapter_to_samples(samples, **kwargs)
        # Optionally split the arrays along the last axis.
        if split:
            raise NotImplementedError("split=True is currently not supported for `PointApproximator`.")

        # Squeeze sample dictionary if there's only one key-value pair.
        samples = self._squeeze_parametric_score_major_dict(samples)

        return samples

    def log_prob(
        self,
        *,
        data: Mapping[str, np.ndarray],
        **kwargs,
    ) -> np.ndarray | dict[str, np.ndarray]:
        """
        Computes the log-probability of given data under the parametric distribution(s) for given input conditions.

        Parameters
        ----------
        data : dict[str, np.ndarray]
            A dictionary mapping variable names to arrays representing the inference conditions and variables.
        **kwargs
            Additional keyword arguments passed to underlying processing functions.

        Returns
        -------
        log_prob : np.ndarray or dict[str, np.ndarray]
            Log-probabilities of the distribution
            `p(inference_variables | inference_conditions, h(summary_conditions))` for all parametric scoring rules.

            If only one parametric score is available, output is an array of log-probabilities.

            Output is a dictionary if multiple parametric scores are available.
            Then, each key is the name of a score and values are corresponding log-probabilities.

            Log-probabilities have shape (num_datasets,).
        """
        log_prob = super().log_prob(data=data, **kwargs)
        # Squeeze log probabilities dictionary if there's only one key-value pair.
        log_prob = PointApproximator._squeeze_parametric_score_major_dict(log_prob)

        return log_prob

    def _prepare_conditions(self, conditions: Mapping[str, np.ndarray], **kwargs) -> dict[str, Tensor]:
        """Adapts and converts the conditions to tensors."""

        conditions = self.adapter(conditions, strict=False, stage="inference", **kwargs)
        conditions = {k: v for k, v in conditions.items() if k in ContinuousApproximator.SAMPLE_KEYS}

        return keras.tree.map_structure(keras.ops.convert_to_tensor, conditions)

    def _apply_inverse_adapter_to_estimates(
        self, estimates: Mapping[str, Mapping[str, Tensor]], **kwargs
    ) -> dict[str, dict[str, dict[str, np.ndarray]]]:
        """Applies the inverse adapter on each inner element of the _estimate output dictionary."""
        estimates = keras.tree.map_structure(keras.ops.convert_to_numpy, estimates)
        processed = {}
        for score_key, score_val in estimates.items():
            processed[score_key] = {}
            for head_key, estimate in score_val.items():
                if head_key in self.inference_network.scores[score_key].NOT_TRANSFORMING_LIKE_VECTOR_WARNING:
                    logging.warning(
                        f"Estimate '{score_key}.{head_key}' is marked to not transform like a vector. "
                        f"It was treated like a vector by the adapter. Handle '{head_key}' estimates with care."
                    )

                adapted = self.adapter(
                    {"inference_variables": estimate},
                    inverse=True,
                    strict=False,
                    **kwargs,
                )
                processed[score_key][head_key] = adapted
        return processed

    def _apply_inverse_adapter_to_samples(
        self, samples: Mapping[str, Tensor], **kwargs
    ) -> dict[str, dict[str, np.ndarray]]:
        """Applies the inverse adapter to a dictionary of samples."""
        samples = keras.tree.map_structure(keras.ops.convert_to_numpy, samples)
        processed = {}
        for score_key, samples in samples.items():
            processed[score_key] = self.adapter(
                {"inference_variables": samples},
                inverse=True,
                strict=False,
                **kwargs,
            )
        return processed

    @staticmethod
    def _reorder_estimates(
        estimates: Mapping[str, Mapping[str, Mapping[str, np.ndarray]]],
    ) -> dict[str, dict[str, dict[str, np.ndarray]]]:
        """Reorders the nested dictionary so that the inference variable names become the top-level keys."""
        # Grab the variable names from one sample inner dictionary.
        sample_inner = next(iter(next(iter(estimates.values())).values()))
        variable_names = sample_inner.keys()
        reordered = {}
        for variable in variable_names:
            reordered[variable] = {}
            for score_key, inner_dict in estimates.items():
                reordered[variable][score_key] = {inner_key: value[variable] for inner_key, value in inner_dict.items()}
        return reordered

    @staticmethod
    def _squeeze_estimates(
        estimates: Mapping[str, Mapping[str, Mapping[str, np.ndarray]]],
    ) -> dict[str, dict[str, np.ndarray]]:
        """Squeezes each inner estimate dictionary to remove unnecessary nesting."""
        squeezed = {}
        for variable, variable_estimates in estimates.items():
            squeezed[variable] = {
                score_key: squeeze_inner_estimates_dict(inner_estimate)
                for score_key, inner_estimate in variable_estimates.items()
            }
        return squeezed

    @staticmethod
    def _squeeze_parametric_score_major_dict(samples: Mapping[str, np.ndarray]) -> np.ndarray or dict[str, np.ndarray]:
        """Squeezes the dictionary to just the value if there is only one key-value pair."""
        if len(samples) == 1:
            return next(iter(samples.values()))  # Extract and return the only item's value
        return samples

    def _estimate(
        self,
        inference_conditions: Tensor = None,
        summary_variables: Tensor = None,
        **kwargs,
    ) -> dict[str, dict[str, Tensor]]:
        if self.summary_network is None:
            if summary_variables is not None:
                raise ValueError("Cannot use summary variables without a summary network.")
        else:
            if summary_variables is None:
                raise ValueError("Summary variables are required when a summary network is present.")

            summary_outputs = self.summary_network(
                summary_variables, **filter_kwargs(kwargs, self.summary_network.call)
            )

            if inference_conditions is None:
                inference_conditions = summary_outputs
            else:
                inference_conditions = keras.ops.concatenate([inference_conditions, summary_outputs], axis=1)

        return self.inference_network(
            conditions=inference_conditions,
            **filter_kwargs(kwargs, self.inference_network.call),
        )



import keras

from bayesflow.types import Tensor
from bayesflow.utils import layer_kwargs
from bayesflow.utils.serialization import serializable

from .mab import MultiHeadAttentionBlock


@serializable
class InducedSetAttentionBlock(keras.Layer):
    """Implements the ISAB block from [1] which represents learnable self-attention specifically
    designed to deal with large sets via a learnable set of "inducing points".

    [1] Lee, J., Lee, Y., Kim, J., Kosiorek, A., Choi, S., & Teh, Y. W. (2019).
        Set transformer: A framework for attention-based permutation-invariant neural networks.
        In International conference on machine learning (pp. 3744-3753). PMLR.
    """

    def __init__(
        self,
        num_inducing_points: int,
        embed_dim: int = 64,
        num_heads: int = 4,
        dropout: float = 0.05,
        mlp_depth: int = 2,
        mlp_width: int = 128,
        mlp_activation: str = "gelu",
        kernel_initializer: str = "lecun_normal",
        use_bias: bool = True,
        layer_norm: bool = True,
        **kwargs,
    ):
        """Creates a self-attention attention block with inducing points (ISAB) which will typically
        be used as part of a set transformer architecture according to [1].

        Parameters
        ----------
        num_inducing_points : int, optional
            The number of inducing points for set-based dimensionality reduction.
        embed_dim : int, optional
            Dimensionality of the embedding space, by default 64.
        num_heads : int, optional
            Number of attention heads, by default 4.
        dropout : float, optional
            Dropout rate applied to attention and MLP layers, by default 0.05.
        mlp_depth : int, optional
            Number of layers in the feedforward MLP block, by default 2.
        mlp_width : int, optional
            Width of each hidden layer in the MLP block, by default 128.
        mlp_activation : str, optional
            Activation function used in the MLP block, by default "gelu".
        kernel_initializer : str, optional
            Initializer for kernel weights, by default "lecun_normal".
        use_bias : bool, optional
            Whether to include bias terms in dense layers, by default True.
        layer_norm : bool, optional
            Whether to apply layer normalization before and after attention, by default True.
        **kwargs : dict
            Additional keyword arguments passed to the Keras Layer base class.
        """

        super().__init__(**layer_kwargs(kwargs))

        self.num_inducing_points = num_inducing_points
        self.inducing_points = self.add_weight(
            shape=(self.num_inducing_points, embed_dim),
            initializer="glorot_uniform",
            trainable=True,
        )
        mab_kwargs = dict(
            embed_dim=embed_dim,
            num_heads=num_heads,
            dropout=dropout,
            mlp_depth=mlp_depth,
            mlp_width=mlp_width,
            mlp_activation=mlp_activation,
            kernel_initializer=kernel_initializer,
            use_bias=use_bias,
            layer_norm=layer_norm,
        )
        self.mab0 = MultiHeadAttentionBlock(**mab_kwargs)
        self.mab1 = MultiHeadAttentionBlock(**mab_kwargs)

    def call(self, input_set: Tensor, training: bool = False, **kwargs) -> Tensor:
        """Performs the forward pass through the self-attention layer.

        Parameters
        ----------
        input_set  : Tensor (e.g., np.ndarray, tf.Tensor, ...)
            Input of shape (batch_size, set_size, input_dim)
            Since this is self-attention, the input set is used
            as a query (Q), key (K), and value (V)
        training   : boolean, optional (default - True)
            Passed to the optional internal dropout and spectral normalization
            layers to distinguish between train and test time behavior.
        **kwargs   : dict, optional (default - {})
            Additional keyword arguments passed to the internal attention layer,
            such as ``attention_mask`` or ``return_attention_scores``

        Returns
        -------
        out : Tensor
            Output of shape (batch_size, set_size, input_dim)
        """

        batch_size = keras.ops.shape(input_set)[0]
        inducing_points_expanded = keras.ops.expand_dims(self.inducing_points, axis=0)
        inducing_points_tiled = keras.ops.tile(inducing_points_expanded, [batch_size, 1, 1])
        h = self.mab0(inducing_points_tiled, input_set, training=training, **kwargs)
        return self.mab1(input_set, h, training=training, **kwargs)



from collections.abc import Callable, Sequence
import numpy as np

from bayesflow.types import Shape
from bayesflow.utils import tree_concatenate
from bayesflow.utils.decorators import allow_batch_size

from bayesflow.utils import numpy_utils as npu

from types import FunctionType

from .simulator import Simulator
from .lambda_simulator import LambdaSimulator


class ModelComparisonSimulator(Simulator):
    """Wraps a sequence of simulators for use with a model comparison approximator."""

    def __init__(
        self,
        simulators: Sequence[Simulator],
        p: Sequence[float] = None,
        logits: Sequence[float] = None,
        use_mixed_batches: bool = True,
        shared_simulator: Simulator | Callable[[Sequence[int]], dict[str, any]] = None,
    ):
        """
        Initialize a multimodel simulator that can generate data for mixture / model comparison problems.

        Parameters
        ----------
        simulators : Sequence[Simulator]
            A sequence of simulator instances, each representing a different model.
        p : Sequence[float], optional
            A sequence of probabilities associated with each simulator. Must sum to 1.
            Mutually exclusive with `logits`.
        logits : Sequence[float], optional
            A sequence of logits corresponding to model probabilities. Mutually exclusive with `p`.
            If neither `p` nor `logits` is provided, defaults to uniform logits.
        use_mixed_batches : bool, optional
            If True, samples in a batch are drawn from different models. If False, the entire batch
            is drawn from a single model chosen according to the model probabilities. Default is True.
        shared_simulator : Simulator or Callable, optional
            A shared simulator whose outputs are passed to all model simulators. If a function is
            provided, it is wrapped in a `LambdaSimulator` with batching enabled.
        """
        self.simulators = simulators

        if isinstance(shared_simulator, FunctionType):
            shared_simulator = LambdaSimulator(shared_simulator, is_batched=True)
        self.shared_simulator = shared_simulator

        match logits, p:
            case (None, None):
                logits = [0.0] * len(simulators)
            case (None, logits):
                logits = logits
            case (p, None):
                p = np.array(p)
                if not np.isclose(np.sum(p), 1.0):
                    raise ValueError("Probabilities must sum to 1.")
                logits = np.log(p) - np.log(1 - p)
            case _:
                raise ValueError("Received conflicting arguments. At most one of `p` or `logits` must be provided.")

        if len(logits) != len(simulators):
            raise ValueError(f"Length of logits ({len(logits)}) must match number of simulators ({len(simulators)}).")

        self.logits = logits
        self.use_mixed_batches = use_mixed_batches

    @allow_batch_size
    def sample(self, batch_shape: Shape, **kwargs) -> dict[str, np.ndarray]:
        """
        Sample from the model comparison simulator.

        Parameters
        ----------
        batch_shape : Shape
            The shape of the batch to sample. Typically, a tuple indicating the number of samples,
            but the user can also supply an int.
        **kwargs
            Additional keyword arguments passed to each simulator. These may include outputs from
            the shared simulator.

        Returns
        -------
        data : dict of str to np.ndarray
            A dictionary containing the sampled outputs. Includes:
              - outputs from the selected simulator(s)
              - optionally, outputs from the shared simulator
              - "model_indices": a one-hot encoded array indicating the model origin of each sample
        """
        data = {}
        if self.shared_simulator:
            data |= self.shared_simulator.sample(batch_shape, **kwargs)

        softmax_logits = npu.softmax(self.logits)
        num_models = len(self.simulators)

        # generate data randomly from each model (slower)
        if self.use_mixed_batches:
            model_counts = np.random.multinomial(n=batch_shape[0], pvals=softmax_logits)

            sims = [
                simulator.sample(n, **(kwargs | data)) for simulator, n in zip(self.simulators, model_counts) if n > 0
            ]
            sims = tree_concatenate(sims, numpy=True)
            data |= sims

            model_indices = np.repeat(np.eye(num_models, dtype="int32"), model_counts, axis=0)

        # draw one model index for the whole batch (faster)
        else:
            model_index = np.random.choice(num_models, p=softmax_logits)

            data = self.simulators[model_index].sample(batch_shape, **(kwargs | data))
            model_indices = npu.one_hot(np.full(batch_shape, model_index, dtype="int32"), num_models)

        return data | {"model_indices": model_indices}



"""
Custom types used for type annotations.

.. currentmodule:: bayesflow.types

.. autodata:: Shape
.. autodata:: Tensor
"""

from .shape import Shape
from .tensor import Tensor

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])



from collections.abc import Callable
import keras

from bayesflow.types import Tensor


def vjp(f: Callable[[Tensor], Tensor], x: Tensor, return_output: bool = False):
    """Compute the vector-Jacobian product of f at x."""
    match keras.backend.backend():
        case "jax":
            import jax

            fx, _vjp_fn = jax.vjp(f, x)

            def vjp_fn(projector):
                return _vjp_fn(projector)[0]
        case "tensorflow":
            import tensorflow as tf

            with tf.GradientTape(persistent=True) as tape:
                tape.watch(x)
                fx = f(x)

            def vjp_fn(projector):
                return tape.gradient(fx, x, projector)
        case "torch":
            import torch

            x = keras.ops.copy(x)
            x.requires_grad_(True)

            with torch.enable_grad():
                fx = f(x)

            def vjp_fn(projector):
                return torch.autograd.grad(fx, x, projector, retain_graph=True)[0]
        case other:
            raise NotImplementedError(f"Cannot build a vjp function for backend '{other}'.")

    if return_output:
        return fx, vjp_fn

    return vjp_fn



import numpy as np

from bayesflow.types import Shape
from bayesflow.utils import batched_call, tree_stack
from bayesflow.utils.decorators import allow_batch_size

from ..simulator import Simulator


class BenchmarkSimulator(Simulator):
    @allow_batch_size
    def sample(self, batch_shape: Shape, **kwargs) -> dict[str, np.ndarray]:
        """Runs simulated benchmark and returns `batch_size` parameter
        and observation batches

        Parameters
        ----------
        batch_shape: tuple
            Number of parameter-observation batches to simulate.

        Returns
        -------
        dict[str, np.ndarray]: simulated parameters and observables
            with shapes (`batch_size`, ...)
        """

        data = batched_call(self, batch_shape, kwargs=kwargs, flatten=True)
        data = tree_stack(data, axis=0, numpy=True)
        return data

    def prior(self) -> np.ndarray:
        raise NotImplementedError

    def observation_model(self, params: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def __call__(self, **kwargs) -> dict[str, np.ndarray]:
        prior_draws = self.prior()
        observables = self.observation_model(prior_draws)
        return dict(parameters=prior_draws.astype(np.float32), observables=observables.astype(np.float32))



import numpy as np

import keras
from keras import ops

from bayesflow.types import Tensor
from bayesflow.utils.serialization import serializable


@serializable
class FourierEmbedding(keras.Layer):
    """Implements a Fourier projection with normally distributed frequencies."""

    def __init__(
        self,
        embed_dim: int = 8,
        scale: float = 1.0,
        initializer: str = "random_normal",
        trainable: bool = True,
        include_identity: bool = True,
        **kwargs,
    ):
        """
        Initialize a Fourier projection instance with normally distributed frequencies.

        Parameter
        ---------
        embed_dim        : int (even)
            Dimensionality of the Fourier projection. The resulting embedding
            has dimensionality `embed_dim + 1` if `include_identity` is set to True.
        scale            : float, optional (default - 1.0)
            Scaling factor for the frequencies.
        initializer      : str, optional (default - "random_normal")
            Method for initializing the projection weights.
        trainable        : bool, optional (default - True)
            If True, the projection weights are trainable.
        include_identity : bool, optional (default - True)
            If True, adds an identity mapping component to the embedding.
        """

        super().__init__(**kwargs)

        if embed_dim % 2 != 0:
            raise ValueError(f"Embedding dimension must be even, but is {embed_dim}.")

        self.w = self.add_weight(initializer=initializer, shape=(embed_dim // 2,), trainable=trainable)
        self.scale = scale
        self.embed_dim = embed_dim
        self.include_identity = include_identity

    def call(self, t: Tensor) -> Tensor:
        """Embeds the one-dimensional time scalar into a higher-dimensional Fourier embedding.

        Parameters
        ----------
        t   : Tensor of shape (batch_size, 1)
            vector of times

        Returns
        -------
        emb : Tensor
            Embedding of shape (batch_size, fourier_emb_dim) if `include_identity`
            is False, else (batch_size, fourier_emb_dim+1)
        """
        proj = t * self.w[None, :] * 2 * np.pi * self.scale
        if self.include_identity:
            emb = ops.concatenate([t, ops.sin(proj), ops.cos(proj)], axis=-1)
        else:
            emb = ops.concatenate([ops.sin(proj), ops.cos(proj)], axis=-1)
        return emb



import keras
import keras.ops as ops

from bayesflow.networks import MLP
from bayesflow.types import Tensor
from bayesflow.utils import layer_kwargs
from bayesflow.utils.decorators import sanitize_input_shape
from bayesflow.utils.serialization import serializable

from .mab import MultiHeadAttentionBlock


@serializable
class PoolingByMultiHeadAttention(keras.Layer):
    """Implements the pooling with multi-head attention (PMA) block from [1] which represents
    a permutation-invariant encoder for set-based inputs.

    [1] Lee, J., Lee, Y., Kim, J., Kosiorek, A., Choi, S., & Teh, Y. W. (2019).
        Set transformer: A framework for attention-based permutation-invariant neural networks.
        In International conference on machine learning (pp. 3744-3753). PMLR.

    Note: Currently works only on 3D inputs but can easily be expanded by changing
    the internals slightly or using ``keras.layers.TimeDistributed``.
    """

    def __init__(
        self,
        num_seeds: int = 1,
        embed_dim: int = 64,
        num_heads: int = 4,
        seed_dim: int = None,
        dropout: float = 0.05,
        mlp_depth: int = 2,
        mlp_width: int = 128,
        mlp_activation: str = "gelu",
        kernel_initializer: str = "lecun_normal",
        use_bias: bool = True,
        layer_norm: bool = True,
        **kwargs,
    ):
        """
        Creates a PoolingByMultiHeadAttention (PMA) block for permutation-invariant set encoding using
        multi-head attention pooling. Can also be used us a building block for `DeepSet` architectures.

        Parameters
        ----------
        num_seeds : int, optional (default=1)
            Number of seed vectors used for pooling. Acts as the number of summary outputs.
        embed_dim : int, optional (default=64)
            Dimensionality of the embedding space used in the attention mechanism.
        num_heads : int, optional (default=4)
            Number of attention heads in the multi-head attention block.
        seed_dim : int or None, optional (default=None)
            Dimensionality of each seed vector. If None, defaults to `embed_dim`.
        dropout : float, optional (default=0.05)
            Dropout rate applied to attention and MLP layers.
        mlp_depth : int, optional (default=2)
            Number of layers in the feedforward MLP applied before attention.
        mlp_width : int, optional (default=128)
            Number of units in each hidden layer of the MLP.
        mlp_activation : str, optional (default="gelu")
            Activation function used in the MLP.
        kernel_initializer : str, optional (default="lecun_normal")
            Initializer for kernel weights in dense layers.
        use_bias : bool, optional (default=True)
            Whether to include bias terms in dense layers.
        layer_norm : bool, optional (default=True)
            Whether to apply layer normalization before and after attention.
        **kwargs
            Additional keyword arguments passed to the Keras Layer base class.
        """

        super().__init__(**layer_kwargs(kwargs))

        self.mab = MultiHeadAttentionBlock(
            embed_dim=embed_dim,
            num_heads=num_heads,
            dropout=dropout,
            mlp_depth=mlp_depth,
            mlp_width=mlp_width,
            mlp_activation=mlp_activation,
            kernel_initializer=kernel_initializer,
            use_bias=use_bias,
            layer_norm=layer_norm,
        )

        self.seed_vector = self.add_weight(
            shape=(num_seeds, seed_dim if seed_dim is not None else embed_dim),
            initializer="glorot_uniform",
            trainable=True,
        )

        self.feedforward = MLP(
            widths=(mlp_width,) * mlp_depth,
            activation=mlp_activation,
            kernel_initializer=kernel_initializer,
            dropout=dropout,
        )

    def call(self, input_set: Tensor, training: bool = False, **kwargs) -> Tensor:
        """Performs the forward pass through the PMA block.

        Parameters
        ----------
        input_set  : Tensor (e.g., np.ndarray, tf.Tensor, ...)
            Input of shape (batch_size, set_size, input_dim)
            Since this is self-attention, the input set is used
            as a query (Q), key (K), and value (V)
        training   : boolean, optional (default - True)
            Passed to the optional internal dropout and spectral normalization
            layers to distinguish between train and test time behavior.
        **kwargs   : dict, optional (default - {})
            Additional keyword arguments passed to the internal attention layer,
            such as ``attention_mask`` or ``return_attention_scores``

        Returns
        -------
        summary : Tensor
            Output of shape (batch_size, num_seeds * summary_dim)
        """

        set_x_transformed = self.feedforward(input_set, training=training)
        batch_size = ops.shape(input_set)[0]
        seed_vector_expanded = ops.expand_dims(self.seed_vector, axis=0)
        seed_tiled = ops.tile(seed_vector_expanded, [batch_size, 1, 1])
        summaries = self.mab(seed_tiled, set_x_transformed, training=training, **kwargs)
        return ops.reshape(summaries, (ops.shape(summaries)[0], -1))

    @sanitize_input_shape
    def compute_output_shape(self, input_shape):
        return keras.ops.shape(self.call(keras.ops.zeros(input_shape)))



from collections.abc import Sequence
import numpy as np

from bayesflow.utils.serialization import serialize, serializable

from .transform import Transform


@serializable
class Broadcast(Transform):
    """
    Broadcasts arrays or scalars to the shape of a given other array.

    Parameters
    ----------
    keys : sequence of str,
        Input a list of strings, where the strings are the names of data variables.
    to : str
        Name of the data variable to broadcast to.
    expand : str or int or tuple, optional
        Where should new dimensions be added to match the number of dimensions in `to`?
        Can be "left", "right", or an integer or tuple containing the indices of the new dimensions.
        The latter is needed if we want to include a dimension in the middle, which will be required
        for more advanced cases. By default we expand left.
    exclude : int or tuple, optional
        Which dimensions (of the dimensions after expansion) should retain their size,
        rather than being broadcasted to the corresponding dimension size of `to`?
        By default we exclude the last dimension (usually the data dimension) from broadcasting the size.
    squeeze : int or tuple, optional
        Axis to squeeze after broadcasting.

    Notes
    -----
    Important: Do not broadcast to variables that are used as inference variables
    (i.e., parameters to be inferred by the networks). The adapter will work during training
    but then fail during inference because the variable being broadcasted to is not available.

    Examples
    --------
    shape (1, ) array:

    >>> a = np.array((1,))

    shape (2, 3) array:

    >>> b = np.array([[1, 2, 3], [4, 5, 6]])

    shape (2, 2, 3) array:

    >>> c = np.array([[[1, 2, 3], [4, 5, 6]], [[4, 5, 6], [1, 2, 3]]])

    >>> dat = dict(a=a, b=b, c=c)
    >>> bc = bf.adapters.transforms.Broadcast("a", to="b")
    >>> new_dat = bc.forward(dat)
    >>> new_dat["a"].shape
    (2, 1)

    >>> bc = bf.adapters.transforms.Broadcast("a", to="b", exclude=None)
    >>> new_dat = bc.forward(dat)
    >>> new_dat["a"].shape
    (2, 3)

    >>> bc = bf.adapters.transforms.Broadcast("b", to="c", expand=1)
    >>> new_dat = bc.forward(dat)
    >>> new_dat["b"].shape
    (2, 2, 3)

    It is recommended to precede this transform with a :class:`bayesflow.adapters.transforms.ToArray` transform.
    """

    def __init__(
        self,
        keys: Sequence[str],
        *,
        to: str,
        expand: str | int | tuple = "left",
        exclude: int | tuple = -1,
        squeeze: int | tuple = None,
    ):
        super().__init__()
        self.keys = keys
        self.to = to

        if isinstance(expand, int):
            expand = (expand,)

        self.expand = expand

        if isinstance(exclude, int):
            exclude = (exclude,)

        self.exclude = exclude
        self.squeeze = squeeze

    def get_config(self) -> dict:
        config = {
            "keys": self.keys,
            "to": self.to,
            "expand": self.expand,
            "exclude": self.exclude,
            "squeeze": self.squeeze,
        }
        return serialize(config)

    # noinspection PyMethodOverriding
    def forward(self, data: dict[str, np.ndarray], **kwargs) -> dict[str, np.ndarray]:
        target_shape = data[self.to].shape

        data = data.copy()

        for k in self.keys:
            # ensure that .shape is defined
            data[k] = np.asarray(data[k])
            len_diff = len(target_shape) - len(data[k].shape)

            if self.expand == "left":
                data[k] = np.expand_dims(data[k], axis=tuple(np.arange(0, len_diff)))
            elif self.expand == "right":
                data[k] = np.expand_dims(data[k], axis=tuple(-np.arange(1, len_diff + 1)))
            elif isinstance(self.expand, tuple):
                if len(self.expand) is not len_diff:
                    raise ValueError("Length of `expand` must match the length difference of the involed arrays.")
                data[k] = np.expand_dims(data[k], axis=self.expand)

            new_shape = target_shape
            if self.exclude is not None:
                new_shape = np.array(new_shape, dtype=int)
                old_shape = np.array(data[k].shape, dtype=int)
                exclude = list(self.exclude)
                new_shape[exclude] = old_shape[exclude]
                new_shape = tuple(new_shape)

            data[k] = np.broadcast_to(data[k], new_shape)

            if self.squeeze is not None:
                data[k] = np.squeeze(data[k], axis=self.squeeze)

        return data

    # noinspection PyMethodOverriding
    def inverse(self, data: dict[str, np.ndarray], **kwargs) -> dict[str, np.ndarray]:
        # TODO: add inverse
        # we will likely never actually need the inverse broadcasting in practice
        # so adding this method is not high priority
        return data



import numpy as np

from .benchmark_simulator import BenchmarkSimulator


class GaussianMixture(BenchmarkSimulator):
    def __init__(
        self,
        D: int = 2,
        lower_bound: float = -10.0,
        upper_bound: float = 10.0,
        prob: float = 0.5,
        scale_c1: float = 1.0,
        scale_c2: float = 0.1,
        rng: np.random.Generator = None,
    ):
        """Gaussian Mixture simulated benchmark
        See: https://arxiv.org/pdf/2101.04653.pdf, Task T.7

        Important: The parameterization uses scales, so use sqrt(var),
        if you want to be working with variances instead of scales.

        Parameters
        ----------
        D: int, optional, default: 2
            The dimensionality of the mixture model.
        lower_bound: float, optional, default: -10.0
            The lower bound of the uniform prior.
        upper_bound: float, optional, default: 10.0
            The upper bound of the uniform prior.
        prob: float, optional, default: 0.5
            The mixture probability (coefficient).
        scale_c1: float, optional, default: 1.0
            The scale of the first component.
        scale_c2: float, optional, default: 0.1
            The scale of the second component.
        rng: np.random.Generator or None, optional, default: None
            An optional random number generator to use.
        """

        self.D = D
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.prob = prob
        self.scale_c1 = scale_c1
        self.scale_c2 = scale_c2
        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()

    def prior(self):
        """Generates a random draw from a 2-dimensional uniform prior bounded between
        `lower_bound` and `upper_bound` representing the common mean of a 2D Gaussian
        mixture model (GMM).

        Returns
        -------
        params : np.ndarray of shape (D, )
            A single draw from the D-dimensional uniform prior
        """

        return self.rng.uniform(low=self.lower_bound, high=self.upper_bound, size=self.D)

    def observation_model(self, params: np.ndarray):
        """Simulates data from the Gaussian mixture model (GMM) with
        shared location vector. For more details, see

        Parameters
        ----------
        params   : np.ndarray of shape (D, )
            The D-dimensional vector of parameter locations.

        Returns
        -------
        x : np.ndarray of shape (2, )
            The 2D vector generated from the GMM simulator.
        """

        # Draw component index
        idx = self.rng.binomial(n=1, p=self.prob)

        # Draw 2D-Gaussian sample according to component index
        if idx == 0:
            return self.rng.normal(loc=params, scale=self.scale_c1)
        return self.rng.normal(loc=params, scale=self.scale_c2)



from collections.abc import Mapping, Sequence

import numpy as np
import matplotlib.pyplot as plt

from ...utils.plot_utils import prepare_plot_data, add_titles_and_labels, prettify_subplots
from ...utils.ecdf import simultaneous_ecdf_bands
from ...utils.ecdf.ranks import fractional_ranks, distance_ranks


def calibration_ecdf(
    estimates: Mapping[str, np.ndarray] | np.ndarray,
    targets: Mapping[str, np.ndarray] | np.ndarray,
    variable_keys: Sequence[str] = None,
    variable_names: Sequence[str] = None,
    difference: bool = False,
    stacked: bool = False,
    rank_type: str | np.ndarray = "fractional",
    figsize: Sequence[float] = None,
    label_fontsize: int = 16,
    legend_fontsize: int = 14,
    legend_location: str = "upper right",
    title_fontsize: int = 18,
    tick_fontsize: int = 12,
    rank_ecdf_color: str = "#132a70",
    fill_color: str = "grey",
    num_row: int = None,
    num_col: int = None,
    **kwargs,
) -> plt.Figure:
    """
    Creates the empirical CDFs for each marginal rank distribution
    and plots it against a uniform ECDF.
    ECDF simultaneous bands are drawn using simulations from the uniform,
    as proposed by [1].

    For models with many parameters, use `stacked=True` to obtain an idea
    of the overall calibration of a posterior approximator.

    To compute ranks based on the Euclidean distance to the origin or a reference, use `rank_type='distance'` (and
    pass a reference array, respectively). This can be used to check the joint calibration of the posterior approximator
    and might show potential biases in the posterior approximation which are not detected by the fractional ranks (e.g.,
    when the prior equals the posterior). This is motivated by [2].

    [1] Säilynoja, T., Bürkner, P. C., & Vehtari, A. (2022). Graphical test
    for discrete uniformity and its applications in goodness-of-fit evaluation
    and multiple sample comparison. Statistics and Computing, 32(2), 1-21.
    https://arxiv.org/abs/2103.10522

    [2] Lemos, Pablo, et al. "Sampling-based accuracy testing of posterior estimators
     for general inference." International Conference on Machine Learning. PMLR, 2023.
     https://proceedings.mlr.press/v202/lemos23a.html

    Parameters
    ----------
    estimates      : np.ndarray of shape (n_data_sets, n_post_draws, n_params)
        The posterior draws obtained from n_data_sets
    targets     : np.ndarray of shape (n_data_sets, n_params)
        The prior draws obtained for generating n_data_sets
    difference        : bool, optional, default: False
        If `True`, plots the ECDF difference.
        Enables a more dynamic visualization range.
    stacked           : bool, optional, default: False
        If `True`, all ECDFs will be plotted on the same plot.
        If `False`, each ECDF will have its own subplot,
        similar to the behavior of `calibration_histogram`.
    rank_type   : str, optional, default: 'fractional'
        If `fractional` (default), the ranks are computed as the fraction
        of posterior samples that are smaller than the prior.
        If `distance`, the ranks are computed as the fraction of posterior
        samples that are closer to a reference points (default here is the origin).
        You can pass a reference array in the same shape as the
        `estimates` array by setting `targets` in the ``ranks_kwargs``.
        This is motivated by [2].
    variable_keys       : list or None, optional, default: None
       Select keys from the dictionaries provided in estimates and targets.
       By default, select all keys.
    variable_names    : list or None, optional, default: None
        The parameter names for nice plot titles.
        Inferred if None. Only relevant if `stacked=False`.
    figsize           : tuple or None, optional, default: None
        The figure size passed to the matplotlib constructor.
        Inferred if None.
    label_fontsize    : int, optional, default: 16
        The font size of the y-label and y-label texts
    legend_fontsize   : int, optional, default: 14
        The font size of the legend text
    title_fontsize    : int, optional, default: 18
        The font size of the title text.
        Only relevant if `stacked=False`
    tick_fontsize     : int, optional, default: 12
        The font size of the axis ticklabels
    rank_ecdf_color   : str, optional, default: '#a34f4f'
        The color to use for the rank ECDFs
    fill_color        : str, optional, default: 'grey'
        The color of the fill arguments.
    num_row           : int, optional, default: None
        The number of rows for the subplots.
        Dynamically determined if None.
    num_col           : int, optional, default: None
        The number of columns for the subplots.
        Dynamically determined if None.
    **kwargs          : dict, optional, default: {}
        Keyword arguments can be passed to control the behavior of
        ECDF simultaneous band computation through the ``ecdf_bands_kwargs``
        dictionary. See `simultaneous_ecdf_bands` for keyword arguments.
        Moreover, additional keyword arguments can be passed to control the behavior of
        the rank computation through the ``ranks_kwargs`` dictionary.

    Returns
    -------
    f : plt.Figure - the figure instance for optional saving

    Raises
    ------
    ShapeError
        If there is a deviation form the expected shapes of `estimates`
        and `targets`.
    ValueError
        If an unknown `rank_type` is passed.
    """

    plot_data = prepare_plot_data(
        estimates=estimates,
        targets=targets,
        variable_keys=variable_keys,
        variable_names=variable_names,
        num_col=num_col,
        num_row=num_row,
        figsize=figsize,
        stacked=stacked,
    )

    estimates = plot_data.pop("estimates")
    targets = plot_data.pop("targets")

    if rank_type == "fractional":
        # Compute fractional ranks
        ranks = fractional_ranks(estimates, targets)
    elif rank_type == "distance":
        # Compute ranks based on distance to the origin
        ranks = distance_ranks(estimates, targets, stacked=stacked, **kwargs.pop("ranks_kwargs", {}))
    else:
        raise ValueError(f"Unknown rank type: {rank_type}. Use 'fractional' or 'distance'.")

    # Plot individual ecdf of parameters
    for j in range(ranks.shape[-1]):
        xx = np.repeat(np.sort(ranks[:, j]), 2)
        xx = np.pad(xx, (1, 1), constant_values=(0, 1))
        yy = np.linspace(0, 1, num=xx.shape[-1] // 2)
        yy = np.repeat(yy, 2)

        # Difference, if specified
        if difference:
            yy -= xx

        if stacked:
            if j == 0:
                if not isinstance(plot_data["axes"], np.ndarray):
                    plot_data["axes"] = np.array([plot_data["axes"]])  # in case of single axis
                plot_data["axes"][0].plot(xx, yy, color=rank_ecdf_color, alpha=0.95, label="Rank ECDFs")
            else:
                plot_data["axes"][0].plot(xx, yy, color=rank_ecdf_color, alpha=0.95)
        else:
            plot_data["axes"].flat[j].plot(xx, yy, color=rank_ecdf_color, alpha=0.95, label="Rank ECDF")

    # Compute uniform ECDF and bands
    alpha, z, L, U = simultaneous_ecdf_bands(estimates.shape[0], **kwargs.pop("ecdf_bands_kwargs", {}))

    # Difference, if specified
    if difference:
        L -= z
        U -= z
        ylab = "ECDF Difference"
    else:
        ylab = "ECDF"

    # Add simultaneous bounds
    if not stacked:
        titles = plot_data["variable_names"]
    elif rank_type in ["distance", "random"]:
        titles = ["Joint ECDFs"]
    else:
        titles = ["Stacked ECDFs"]

    for ax, title in zip(plot_data["axes"].flat, titles):
        ax.fill_between(z, L, U, color=fill_color, alpha=0.2, label=rf"{int((1 - alpha) * 100)}$\%$ Confidence Bands")
        ax.legend(fontsize=legend_fontsize, loc=legend_location)
        ax.set_title(title, fontsize=title_fontsize)

    prettify_subplots(plot_data["axes"], num_subplots=plot_data["num_variables"], tick_fontsize=tick_fontsize)

    add_titles_and_labels(
        plot_data["axes"],
        plot_data["num_row"],
        plot_data["num_col"],
        xlabel=f"{rank_type.capitalize()} rank statistic",
        ylabel=ylab,
        label_fontsize=label_fontsize,
    )

    plot_data["fig"].tight_layout()
    return plot_data["fig"]



from keras import ops

from bayesflow.types import Tensor

# hard coded from ops.logspace(-6, 6, 11)
# to avoid pytorch errors/warnings if you want to use MPS
default_scales = ops.convert_to_tensor(
    [
        1.0000e-06,
        1.5849e-05,
        2.5119e-04,
        3.9811e-03,
        6.3096e-02,
        1.0000e00,
        1.5849e01,
        2.5119e02,
        3.9811e03,
        6.3096e04,
        1.0000e06,
    ]
)


def gaussian(x: Tensor, y: Tensor, scales: Tensor = default_scales) -> Tensor:
    """Computes a mixture of Gaussian radial basis functions (RBFs) between the samples of x and y.

    Parameters
    ----------
    x       :  Tensor of shape (num_draws_x, num_features)
        Comprises `num_draws_x` Random draws from the "source" distribution `P`.
    y       :  Tensor of shape (num_draws_y, num_features)
        Comprises `num_draws_y` Random draws from the "source" distribution `Q`.
    scales  : Tensor, optional (default - default_scales)
        List which denotes the widths of each of the Gaussians in the mixture.

    Returns
    -------
    kernel_matrix : Tensor of shape (num_draws_x, num_draws_y)
        The kernel matrix between pairs from `x ~ P` and `y ~ Q`.
    """
    beta = 1.0 / (2.0 * scales[..., None])
    dist = x[..., None] - ops.transpose(y)
    dist = ops.transpose(ops.norm(dist, ord=2, axis=1))
    s = ops.matmul(beta, ops.reshape(dist, newshape=(1, -1)))
    return ops.reshape(ops.sum(ops.exp(-s), axis=0), newshape=ops.shape(dist))


def inverse_multiquadratic(x: Tensor, y: Tensor, scales: Tensor = default_scales) -> Tensor:
    """Computes a mixture of inverse multiquadratic RBFs between the samples of x and y.

    Parameters
    ----------
    x       :  Tensor of shape (num_draws_x, num_features)
        Comprises `num_draws_x` Random draws from the "source" distribution `P`.
    y       :  Tensor of shape (num_draws_y, num_features)
        Comprises `num_draws_y` Random draws from the "source" distribution `Q`.
    scales  : Tensor, optional (default - default_scales)
        List which denotes multiple scales for the IM-RBF kernel mixture.

    Returns
    -------
    kernel_matrix : Tensor of shape (num_draws_x, num_draws_y)
        The kernel matrix between pairs from `x ~ P` and `y ~ Q`.
    """
    dist = ops.expand_dims(ops.sum((x[:, None, :] - y[None, :, :]) ** 2, axis=-1), axis=-1)
    sigmas = ops.expand_dims(scales, axis=0)
    return ops.sum(sigmas / (dist + sigmas), axis=-1)



import keras
from keras import ops

import warnings

from bayesflow.distributions import Distribution
from bayesflow.types import Tensor
from bayesflow.utils import (
    concatenate_valid,
    find_network,
    jacobian,
    jvp,
    model_kwargs,
    vjp,
    weighted_mean,
)
from bayesflow.utils.serialization import deserialize, serializable, serialize

from bayesflow.networks import InferenceNetwork


@serializable
class FreeFormFlow(InferenceNetwork):
    """Implements a dimensionality-preserving Free-form Flow.
    Incorporates ideas from [1-2].

    [1] Draxler, F., Sorrenson, P., Zimmermann, L., Rousselot, A., & Köthe, U. (2024).F
    ree-form flows: Make Any Architecture a Normalizing Flow.
    In International Conference on Artificial Intelligence and Statistics.

    [2] Sorrenson, P., Draxler, F., Rousselot, A., Hummerich, S., Zimmermann, L., &
    Köthe, U. (2024). Lifting Architectural Constraints of Injective Flows.
    In International Conference on Learning Representations.
    """

    ENCODER_MLP_DEFAULT_CONFIG = {
        "widths": (256, 256, 256, 256),
        "activation": "mish",
        "kernel_initializer": "he_normal",
        "residual": True,
        "dropout": 0.05,
        "spectral_normalization": False,
    }

    DECODER_MLP_DEFAULT_CONFIG = {
        "widths": (256, 256, 256, 256),
        "activation": "mish",
        "kernel_initializer": "he_normal",
        "residual": True,
        "dropout": 0.05,
        "spectral_normalization": False,
    }

    def __init__(
        self,
        beta: float = 50.0,
        encoder_subnet: str | keras.Layer = "mlp",
        decoder_subnet: str | keras.Layer = "mlp",
        base_distribution: str | Distribution = "normal",
        hutchinson_sampling: str = "qr",
        encoder_subnet_kwargs: dict = None,
        decoder_subnet_kwargs: dict = None,
        **kwargs,
    ):
        """Creates an instance of a Free-form Flow.

        Parameters
        ----------
        beta                  : float, optional, default: 50.0
        encoder_subnet        : str or type, optional, default: "mlp"
            A neural network type for the flow, will be instantiated using
            encoder_subnet_kwargs. Will be equipped with a projector to ensure
            the correct output dimension and a global skip connection.
        decoder_subnet        : str or type, optional, default: "mlp"
            A neural network type for the flow, will be instantiated using
            decoder_subnet_kwargs. Will be equipped with a projector to ensure
            the correct output dimension and a global skip connection.
        base_distribution     : str, optional, default: "normal"
            The latent distribution
        hutchinson_sampling   : str, optional, default: "qr
            One of `["sphere", "qr"]`. Select the sampling scheme for the
            vectors of the Hutchinson trace estimator.
        **kwargs              : dict, optional, default: {}
            Additional keyword arguments
        """
        super().__init__(base_distribution, **kwargs)

        if encoder_subnet_kwargs or decoder_subnet_kwargs:
            warnings.warn(
                "Using `subnet_kwargs` is deprecated."
                "Instead, instantiate the network yourself and pass the arguments directly.",
                DeprecationWarning,
            )

        encoder_subnet_kwargs = encoder_subnet_kwargs or {}
        decoder_subnet_kwargs = decoder_subnet_kwargs or {}

        if encoder_subnet == "mlp":
            encoder_subnet_kwargs = FreeFormFlow.ENCODER_MLP_DEFAULT_CONFIG.copy() | encoder_subnet_kwargs

        if decoder_subnet == "mlp":
            decoder_subnet_kwargs = FreeFormFlow.DECODER_MLP_DEFAULT_CONFIG.copy() | decoder_subnet_kwargs

        self.encoder_subnet = find_network(encoder_subnet, **encoder_subnet_kwargs)
        self.encoder_projector = keras.layers.Dense(
            units=None, bias_initializer="zeros", kernel_initializer="zeros", name="encoder_projector"
        )

        self.decoder_subnet = find_network(decoder_subnet, **decoder_subnet_kwargs)
        self.decoder_projector = keras.layers.Dense(
            units=None, bias_initializer="zeros", kernel_initializer="zeros", name="decoder_projector"
        )

        self.hutchinson_sampling = hutchinson_sampling
        self.beta = beta

        self.seed_generator = keras.random.SeedGenerator()

    # noinspection PyMethodOverriding
    def build(self, xz_shape, conditions_shape=None):
        if self.built:
            # building when the network is already built can cause issues with serialization
            # see https://github.com/keras-team/keras/issues/21147
            return

        self.base_distribution.build(xz_shape)

        self.encoder_projector.units = xz_shape[-1]
        self.decoder_projector.units = xz_shape[-1]

        # construct input shape for subnet and subnet projector
        input_shape = list(xz_shape)

        if conditions_shape is not None:
            input_shape[-1] += conditions_shape[-1]

        input_shape = tuple(input_shape)

        self.encoder_subnet.build(input_shape)
        self.decoder_subnet.build(input_shape)

        output_shape = self.encoder_subnet.compute_output_shape(input_shape)
        self.encoder_projector.build(output_shape)

        output_shape = self.decoder_subnet.compute_output_shape(input_shape)
        self.decoder_projector.build(output_shape)

    def _forward(
        self, x: Tensor, conditions: Tensor = None, density: bool = False, training: bool = False, **kwargs
    ) -> Tensor | tuple[Tensor, Tensor]:
        if density:
            z, jac = jacobian(
                lambda inp: self.encode(inp, conditions=conditions, training=training, **kwargs), x, return_output=True
            )
            log_det = keras.ops.logdet(jac)

            log_density = self.base_distribution.log_prob(z) + log_det
            return z, log_density

        z = self.encode(x, conditions, training=training, **kwargs)
        return z

    def _inverse(
        self, z: Tensor, conditions: Tensor = None, density: bool = False, training: bool = False, **kwargs
    ) -> Tensor | tuple[Tensor, Tensor]:
        if density:
            x, jac = jacobian(
                lambda inp: self.decode(inp, conditions=conditions, training=training, **kwargs), z, return_output=True
            )
            log_det = keras.ops.logdet(jac)

            log_density = self.base_distribution.log_prob(z) - log_det
            return x, log_density

        x = self.decode(z, conditions, training=training, **kwargs)
        return x

    def encode(self, x: Tensor, conditions: Tensor = None, training: bool = False, **kwargs) -> Tensor:
        if conditions is None:
            inp = x
        else:
            inp = concatenate_valid([x, conditions], axis=-1)
        network_out = self.encoder_projector(
            self.encoder_subnet(inp, training=training, **kwargs), training=training, **kwargs
        )
        return network_out + x

    def decode(self, z: Tensor, conditions: Tensor = None, training: bool = False, **kwargs) -> Tensor:
        if conditions is None:
            inp = z
        else:
            inp = concatenate_valid([z, conditions], axis=-1)
        network_out = self.decoder_projector(
            self.decoder_subnet(inp, training=training, **kwargs), training=training, **kwargs
        )
        return network_out + z

    def _sample_v(self, x):
        batch_size = ops.shape(x)[0]
        total_dim = ops.shape(x)[-1]
        match self.hutchinson_sampling:
            case "qr":
                # Use QR decomposition as described in [2]
                v_raw = keras.random.normal((batch_size, total_dim, 1), dtype=ops.dtype(x), seed=self.seed_generator)
                q = ops.reshape(ops.qr(v_raw)[0], ops.shape(x))
                v = q * ops.sqrt(total_dim)
            case "sphere":
                # Sample from sphere with radius sqrt(total_dim), as implemented in [1]
                v_raw = keras.random.normal((batch_size, total_dim), dtype=ops.dtype(x), seed=self.seed_generator)
                v = v_raw * ops.sqrt(total_dim) / ops.sqrt(ops.sum(v_raw**2, axis=-1, keepdims=True))
            case _:
                raise ValueError(f"{self.hutchinson_sampling} is not a valid value for hutchinson_sampling.")
        return v

    def compute_metrics(
        self, x: Tensor, conditions: Tensor = None, sample_weight: Tensor = None, stage: str = "training"
    ) -> dict[str, Tensor]:
        base_metrics = super().compute_metrics(x, conditions=conditions, sample_weight=sample_weight, stage=stage)
        # sample random vector
        v = self._sample_v(x)

        def encode(x):
            return self.encode(x, conditions, training=stage == "training")

        def decode(z):
            return self.decode(z, conditions, training=stage == "training")

        # VJP computation
        z, vjp_fn = vjp(encode, x)
        v1 = vjp_fn(v)[0]
        # JVP computation
        x_pred, v2 = jvp(decode, (z,), (v,))

        # equivalent: surrogate = ops.matmul(ops.stop_gradient(v2[:, None]), v1[:, :, None])[:, 0, 0]
        surrogate = ops.sum((ops.stop_gradient(v2) * v1), axis=-1)
        nll = -self.base_distribution.log_prob(z)
        maximum_likelihood_loss = nll - surrogate
        reconstruction_loss = ops.sum((x - x_pred) ** 2, axis=-1)

        losses = maximum_likelihood_loss + self.beta * reconstruction_loss
        loss = weighted_mean(losses, sample_weight)

        return base_metrics | {"loss": loss}

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self):
        base_config = super().get_config()
        base_config = model_kwargs(base_config)

        config = {
            "beta": self.beta,
            "encoder_subnet": self.encoder_subnet,
            "decoder_subnet": self.decoder_subnet,
            "base_distribution": self.base_distribution,
            "hutchinson_sampling": self.hutchinson_sampling,
            # we do not need to store subnet_kwargs
        }

        return base_config | serialize(config)



import keras
from keras.saving import register_keras_serializable as serializable

from bayesflow.utils import layer_kwargs
from bayesflow.utils.decorators import sanitize_input_shape


@serializable(package="links.ordered")
class Ordered(keras.Layer):
    """Activation function to link to a tensor which is monotonously increasing along a specified axis."""

    def __init__(self, axis: int, anchor_index: int, **kwargs):
        super().__init__(**layer_kwargs(kwargs))
        self.axis = axis
        self.anchor_index = anchor_index
        self.group_indices = None

        self.config = {"axis": axis, "anchor_index": anchor_index, **kwargs}

    def get_config(self):
        base_config = super().get_config()
        return base_config | self.config

    def build(self, input_shape):
        super().build(input_shape)

        if self.anchor_index % input_shape[self.axis] == 0 or self.anchor_index == -1:
            raise RuntimeError("Anchor should not be first or last index.")

        self.group_indices = dict(
            below=list(range(0, self.anchor_index)),
            above=list(range(self.anchor_index + 1, input_shape[self.axis])),
        )

    def call(self, inputs):
        # Divide in anchor, below and above
        below_inputs = keras.ops.take(inputs, self.group_indices["below"], axis=self.axis)
        anchor_input = keras.ops.take(inputs, self.anchor_index, axis=self.axis)
        anchor_input = keras.ops.expand_dims(anchor_input, axis=self.axis)
        above_inputs = keras.ops.take(inputs, self.group_indices["above"], axis=self.axis)

        # Apply softplus for positivity and cumulate to ensure ordered quantiles
        below = keras.activations.softplus(below_inputs)
        above = keras.activations.softplus(above_inputs)

        below = anchor_input - keras.ops.flip(keras.ops.cumsum(below, axis=self.axis), self.axis)
        above = anchor_input + keras.ops.cumsum(above, axis=self.axis)

        # Concatenate and reshape back
        x = keras.ops.concatenate([below, anchor_input, above], self.axis)
        return x

    @sanitize_input_shape
    def compute_output_shape(self, input_shape):
        return input_shape



import keras
import numpy as np

from bayesflow.utils import filter_kwargs


class NumpyApproximator(keras.Model):
    # noinspection PyMethodOverriding
    def compute_metrics(self, *args, **kwargs) -> dict[str, np.ndarray]:
        # implemented by each respective architecture
        raise NotImplementedError

    def test_step(self, data: dict[str, any]) -> dict[str, np.ndarray]:
        kwargs = filter_kwargs(data | {"stage": "validation"}, self.compute_metrics)
        return self.compute_metrics(**kwargs)

    def train_step(self, data: dict[str, any]) -> dict[str, np.ndarray]:
        raise NotImplementedError("Numpy backend does not support training.")



"""
A collection of :py:class:`~bayesflow.approximators.Approximator`\ s, which embody the inference task and the
neural network components used to perform it.
"""

from .approximator import Approximator
from .continuous_approximator import ContinuousApproximator
from .point_approximator import PointApproximator
from .model_comparison_approximator import ModelComparisonApproximator

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])



import numpy as np
from scipy.stats import multivariate_t

from .benchmark_simulator import BenchmarkSimulator


class SLCPDistractors(BenchmarkSimulator):
    def __init__(
        self,
        lower_bound: float = -3.0,
        upper_bound: float = 3.0,
        n_obs: int = 4,
        n_dist: int = 46,
        dim: int = 2,
        mu_scale: float = 15.0,
        shape_scale: float = 0.01,
        flatten: bool = True,
        rng: np.random.Generator = None,
    ):
        """SLCP Distractors simulated benchmark
        See: https://arxiv.org/pdf/2101.04653.pdf, Task T.4

        Parameters
        ----------
        lower_bound: float, optional, default: -3.0
            The lower bound of the uniform prior.
        upper_bound: float, optional, default: 3.0
            The upper bound of the uniform prior.
        n_obs: int, optional, default: 4
            The number of observations to generate from the slcp likelihood.
        n_dist: int, optional, default: 46
            The number of distractor to draw from the distractor likelihood.
        dim: int, optional, default: 2
            The dimensionality of each student-t distribution in the mixture.
        mu_scale: float, optional, default: 15.0
            The scale of the zero-centered Gaussian prior from which the mean vector
            of each student-t distribution in the mixture is drawn.
        shape_scale: float, optional, default: 0.01
            The scale of the assumed `np.eye(dim)` shape matrix. The default is chosen to keep
            the scale of the distractors and observations relatively similar.
        flatten: bool, optional, default: True
            A flag to indicate whather a 1D (`flatten=True`) or a 2D (`flatten=False`)
            representation of the simulated data is returned.
        rng: np.random.Generator or None, optional, default: None
            An optional random number generator to use.
        """

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.n_obs = n_obs
        self.n_dist = n_dist
        self.dim = dim
        self.mu_scale = mu_scale
        self.shape_scale = shape_scale
        self.flatten = flatten
        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()

    def _get_random_student_t(self):
        """A helper function to create a "frozen" multivariate student-t distribution of dimensions `dim`.

        Returns
        -------
        student : callable (scipy.stats._multivariate.multivariate_t_frozen)
            The student-t generator.
        """

        # Draw mean and return student-t object
        mu = self.mu_scale * self.rng.normal(size=self.dim)
        return multivariate_t(loc=mu, shape=self.shape_scale, df=2, allow_singular=True, seed=self.rng)

    def _draw_mixture_student_t(self, num_students: int):
        """Helper function to generate `n_draws` random draws from a mixture of `num_students`
        multivariate Student-t distributions.

        Uses the function `get_random_student_t` to create each of the studen-t callable objects.

        Parameters
        ----------
        num_students : int
            The number of multivariate student-t mixture components

        Returns
        -------
        sample : np.ndarray of shape (n_draws, dim)
            The random draws from the mixture of students.
        """

        # Obtain a list of scipy frozen distributions (each will have a different mean)
        students = [self._get_random_student_t() for _ in range(num_students)]

        # Obtain the sample of n_draws from the mixture and return
        sample = [students[self.rng.integers(low=0, high=num_students)].rvs() for _ in range(self.n_dist)]
        return np.array(sample)

    def prior(self):
        """Generates a random draw from a 5-dimensional uniform prior bounded between
        `lower_bound` and `upper_bound`.

        Returns
        -------
        params : np.ndarray of shape (5, )
            A single draw from the 5-dimensional uniform prior.
        """

        return self.rng.uniform(low=self.lower_bound, high=self.upper_bound, size=5)

    def observation_model(self, params: np.ndarray):
        """Generates data from the SLCP model designed as a benchmark for a simple likelihood
        and a complex posterior due to a non-linear pushforward params -> x. In addition, it
        outputs uninformative distractor data.

        Parameters
        ----------
        params: np.ndarray of shape (params, D)
            The location parameters of the Gaussian likelihood.

        Returns
        -------
        x: np.ndarray of shape (n_obs*2 + n_dist*2, ) if `flatten=True`, otherwise
            np.ndarray of shape (n_obs + n_dist, 2) if `flatten=False`
        """

        # Specify 2D location
        loc = np.array([params[0], params[1]])

        # Specify 2D covariance matrix
        s1 = params[2] ** 2
        s2 = params[3] ** 2
        rho = np.tanh(params[4])
        cov = rho * s1 * s2
        S_param = np.array([[s1**2, cov], [cov, s2**2]])

        # Obtain informative part of the data
        x_info = self.rng.multivariate_normal(loc, S_param, size=self.n_obs)

        # Obtain uninformative part of the data
        x_uninfo = self._draw_mixture_student_t(20)

        # Concatenate informative with uninformative and return
        x = np.concatenate([x_info, x_uninfo], axis=0)
        if self.flatten:
            return x.flatten()
        return x



import keras
import numpy as np

from tests.utils import assert_allclose


def test_build(invertible_layer, random_samples, random_conditions):
    assert invertible_layer.built is False

    invertible_layer(random_samples)

    assert invertible_layer.built is True

    assert invertible_layer.variables, "Layer has no variables."


def test_variable_batch_size(invertible_layer, random_samples, random_conditions):
    # manual build with one batch size
    invertible_layer.build(keras.ops.shape(random_samples))

    # run with another batch size
    batch_sizes = np.random.choice(10, replace=False, size=3)
    for batch_size in batch_sizes:
        new_input = keras.ops.zeros((batch_size,) + keras.ops.shape(random_samples)[1:])
        invertible_layer(new_input)


def test_output_structure(invertible_layer, random_samples, random_conditions):
    output = invertible_layer(random_samples)

    assert isinstance(output, tuple)
    assert len(output) == 2

    forward_output, forward_log_det = output

    assert keras.ops.is_tensor(forward_output)
    assert keras.ops.is_tensor(forward_log_det)


def test_output_shape(invertible_layer, random_samples, random_conditions):
    forward_output, forward_log_det = invertible_layer(random_samples)

    assert keras.ops.shape(forward_output) == keras.ops.shape(random_samples)
    assert keras.ops.shape(forward_log_det) == (keras.ops.shape(random_samples)[0],)

    inverse_output, inverse_log_det = invertible_layer(random_samples, inverse=True)

    assert keras.ops.shape(inverse_output) == keras.ops.shape(random_samples)
    assert keras.ops.shape(inverse_log_det) == (keras.ops.shape(random_samples)[0],)


def test_cycle_consistency(invertible_layer, random_samples, random_conditions):
    # cycle-consistency means the forward and inverse methods are inverses of each other
    forward_output, forward_log_det = invertible_layer(random_samples)
    inverse_output, inverse_log_det = invertible_layer(forward_output, inverse=True)

    assert_allclose(random_samples, inverse_output, atol=1e-6, msg="Samples are not cycle consistent")
    assert_allclose(forward_log_det, -inverse_log_det, atol=1e-6, msg="Log Determinants are not cycle consistent")


def test_jacobian_numerically(invertible_layer, random_samples, random_conditions):
    from bayesflow.utils import jacobian

    forward_output, forward_log_det = invertible_layer(random_samples)

    numerical_forward_jacobian = jacobian(lambda x: invertible_layer(x)[0], random_samples)

    numerical_forward_log_det = keras.ops.logdet(numerical_forward_jacobian)

    assert_allclose(forward_log_det, numerical_forward_log_det, rtol=1e-4, atol=1e-5)

    inverse_output, inverse_log_det = invertible_layer(random_samples, inverse=True)

    numerical_inverse_jacobian = jacobian(lambda z: invertible_layer(z, inverse=True)[0], random_samples)

    numerical_inverse_log_det = keras.ops.logdet(numerical_inverse_jacobian)

    assert_allclose(inverse_log_det, numerical_inverse_log_det, rtol=1e-4, atol=1e-5)



import keras

from .. import logging

from .euclidean import euclidean


def log_sinkhorn(x1, x2, seed: int = None, **kwargs):
    """
    Log-stabilized version of :py:func:`~bayesflow.utils.optimal_transport.sinkhorn.sinkhorn`.
    Significantly slower than the unstabilized version, so use only when you need numerical stability.
    """
    log_plan = log_sinkhorn_plan(x1, x2, **kwargs)
    assignments = keras.random.categorical(keras.ops.exp(log_plan), num_samples=1, seed=seed)
    assignments = keras.ops.squeeze(assignments, axis=1)

    return assignments


def log_sinkhorn_plan(x1, x2, regularization: float = 1.0, rtol=1e-5, atol=1e-8, max_steps=None):
    """
    Log-stabilized version of :py:func:`~bayesflow.utils.optimal_transport.sinkhorn.sinkhorn_plan`.
    Significantly slower than the unstabilized version, so use only when you need numerical stability.
    """
    cost = euclidean(x1, x2)

    log_plan = cost / -(regularization * keras.ops.mean(cost) + 1e-16)

    def contains_nans(plan):
        return keras.ops.any(keras.ops.isnan(plan))

    def is_converged(plan):
        # for convergence, the plan should be doubly stochastic
        conv0 = keras.ops.all(keras.ops.isclose(keras.ops.logsumexp(plan, axis=0), 0.0, rtol=rtol, atol=atol))
        conv1 = keras.ops.all(keras.ops.isclose(keras.ops.logsumexp(plan, axis=1), 0.0, rtol=rtol, atol=atol))
        return conv0 & conv1

    def cond(_, plan):
        # break the while loop if the plan contains nans or is converged
        return ~(contains_nans(plan) | is_converged(plan))

    def body(steps, plan):
        # Sinkhorn-Knopp: repeatedly normalize the transport plan along each dimension
        plan = keras.ops.log_softmax(plan, axis=0)
        plan = keras.ops.log_softmax(plan, axis=1)

        return steps + 1, plan

    steps = 0
    steps, log_plan = keras.ops.while_loop(cond, body, (steps, log_plan), maximum_iterations=max_steps)

    def do_nothing():
        pass

    def log_steps():
        msg = "Log-Sinkhorn-Knopp converged after {} steps."

        logging.debug(msg, steps)

    def warn_convergence():
        msg = "Log-Sinkhorn-Knopp did not converge after {} steps."

        logging.warning(msg, max_steps)

    def warn_nans():
        msg = "Log-Sinkhorn-Knopp produced NaNs after {} steps."
        logging.warning(msg, steps)

    keras.ops.cond(contains_nans(log_plan), warn_nans, do_nothing)
    keras.ops.cond(is_converged(log_plan), log_steps, warn_convergence)

    return log_plan



from typing import Sequence

import numpy as np

from bayesflow.types import Shape
from bayesflow.utils.decorators import allow_batch_size


class Workflow:
    def make_simulator(self, simulators_list: Sequence[callable], meta_fn: callable):
        raise NotImplementedError("Method must be implemented by caller.")

    @allow_batch_size
    def simulate(self, batch_shape: Shape, **kwargs) -> dict[str, np.ndarray]:
        raise NotImplementedError("Method must be implemented by caller.")

    def build_graph(self, *args, **kwargs):
        raise NotImplementedError("Method must be implemented by caller.")

    def fit(self, **kwargs):
        raise NotImplementedError("Method must be implemented by caller.")

    def sample(self, **kwargs):
        raise NotImplementedError("Method must be implemented by caller.")

    def log_prob(self, **kwargs):
        raise NotImplementedError("Method must be implemented by caller.")

    def plot_diagnostics(self, **kwargs):
        raise NotImplementedError("Method must be implemented by caller.")

    def compute_diagnostics(self, **kwargs):
        raise NotImplementedError("Method must be implemented by caller.")



import numpy as np
import pandas as pd

from bayesflow.utils.serialization import serializable

from .transform import Transform


@serializable
class ToDict(Transform):
    """Convert non-dict batches (e.g., pandas.DataFrame) to dict batches"""

    @classmethod
    def from_config(cls, config: dict, custom_objects=None):
        return cls()

    def get_config(self) -> dict:
        return {}

    def forward(self, data, **kwargs) -> dict[str, np.ndarray]:
        data = dict(data)

        for key, value in data.items():
            if isinstance(value, pd.Series):
                if value.dtype == "object":
                    value = value.astype("category")

                if value.dtype == "category":
                    value = pd.get_dummies(value)

                value = np.asarray(value).astype("float32", copy=False)

            data[key] = value

        return data

    def inverse(self, data: dict[str, np.ndarray], **kwargs) -> dict[str, np.ndarray]:
        # non-invertible transform
        return data



import numpy as np
import keras
import pytest


@pytest.fixture()
def batch_size():
    return 16


@pytest.fixture()
def num_variables():
    return 10


@pytest.fixture()
def generic_preactivation(batch_size):
    return keras.ops.ones((batch_size, 6))


@pytest.fixture()
def ordered():
    from bayesflow.links import Ordered

    return Ordered(axis=1, anchor_index=2)


@pytest.fixture()
def ordered_quantiles():
    from bayesflow.links import OrderedQuantiles

    return OrderedQuantiles()


@pytest.fixture()
def positive_definite():
    from bayesflow.links import PositiveDefinite

    return PositiveDefinite()


@pytest.fixture()
def linear():
    return keras.layers.Activation("linear")


@pytest.fixture(params=["ordered", "ordered_quantiles", "positive_definite", "linear"], scope="function")
def link(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def num_quantiles():
    return 19


@pytest.fixture()
def quantiles_np(num_quantiles):
    return np.linspace(0, 1, num_quantiles + 2)[1:-1]


@pytest.fixture()
def quantiles_py(quantiles_np):
    return list(quantiles_np)


@pytest.fixture()
def quantiles_keras(quantiles_np):
    return keras.ops.convert_to_tensor(quantiles_np)


@pytest.fixture()
def none():
    return None


@pytest.fixture(params=["quantiles_np", "quantiles_py", "quantiles_keras", "none"], scope="function")
def quantiles(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def unordered(batch_size, num_quantiles, num_variables):
    return keras.random.normal((batch_size, num_quantiles, num_variables))



import logging
import os
import shutil
import sys
from pathlib import Path

try:
    from sphinx_polyversion.git import Git

    USE_POLYVERSION = True
except ImportError:
    USE_POLYVERSION = False

logging.basicConfig()
logger = logging.getLogger("pre-build.py")
logger.setLevel(logging.DEBUG)


def copy_files(sourcedir):
    basedir = Path(os.path.abspath(sourcedir)).parent.parent
    logger.info(f"Base directory is '{basedir}'")
    logger.info(f"Documentation source directory is '{sourcedir}'")

    # copy examples
    examples_src = os.path.join(basedir, "examples")
    examples_dst = os.path.join(sourcedir, "_examples")
    if os.path.exists(examples_src):
        logger.info("Copying examples")
        shutil.copytree(examples_src, examples_dst, dirs_exist_ok=True)
    examples_in_progress = os.path.join(examples_dst, "in_progress")
    if os.path.exists(examples_in_progress):
        shutil.rmtree(examples_in_progress)
    # copy contributing and installation
    contributing_src = os.path.join(basedir, "CONTRIBUTING.md")
    contributing_dst = os.path.join(sourcedir, "contributing.md")
    if os.path.exists(contributing_src):
        shutil.copy2(contributing_src, contributing_dst)
    installation_src = os.path.join(basedir, "INSTALL.rst")
    installation_dst = os.path.join(sourcedir, "installation.rst")
    if os.path.exists(installation_src):
        shutil.copy2(installation_src, installation_dst)
    # create empty bibtex file if none exists
    bibtex_path = os.path.join(sourcedir, "references.bib")
    if not os.path.exists(bibtex_path):
        open(bibtex_path, "a").close()


def patch_conf(sourcedir):
    if USE_POLYVERSION:
        root = Git.root(Path(__file__).parent)
    else:
        root = str(Path(os.path.abspath(sourcedir)).parent.parent)
    sourcedir = Path(sourcedir)
    cursrc = Path(root) / "docsrc" / "source"
    if os.path.abspath(cursrc) == os.path.abspath(sourcedir):
        return
    # copy the configuration file: shared for all versions
    conf_src = Path(cursrc) / "conf.py"
    conf_dst = Path(sourcedir) / "conf.py"
    if conf_src.exists():
        logger.info("Overwriting old conf.py with current conf.py")
        shutil.copy2(conf_src, conf_dst)
    # copy HTML and CSS for versioning sidebar
    copy_rel_paths = [
        "_static/bayesflow_hor_dark.png",
        "_static/bayesflow_hor.png",
        "_static/custom.css",
        "sphinxext/override_pst_pagetoc.py",
        "sphinxext/adapt_autodoc_docstring.py",
    ]
    for path in copy_rel_paths:
        srcfile = cursrc / path
        dstfile = sourcedir / path
        if srcfile.exists():
            os.makedirs(dstfile.parent, exist_ok=True)
            shutil.copy2(srcfile, dstfile)


if __name__ == "__main__":
    logger.info("Running pre-build script")  # move files around if necessary
    sourcedir = sys.argv[1]
    copy_files(sourcedir)
    patch_conf(sourcedir)



import numpy as np
from scipy import special


def inverse_sigmoid(x: np.ndarray) -> np.ndarray:
    """Inverse of the sigmoid function."""
    return np.log(x) - np.log1p(-x)


def inverse_shifted_softplus(
    x: np.ndarray, shift: float = np.log(np.e - 1), beta: float = 1.0, threshold: float = 20.0
) -> np.ndarray:
    """Inverse of the shifted softplus function."""
    return inverse_softplus(x, beta=beta, threshold=threshold) - shift


def inverse_softplus(x: np.ndarray, beta: float = 1.0, threshold: float = 20.0) -> np.ndarray:
    """Numerically stabilized inverse softplus function."""
    with np.errstate(over="ignore"):
        expm1_x = np.expm1(x)
    return np.where(beta * x > threshold, x, np.log(beta * expm1_x) / beta)


def one_hot(indices: np.ndarray, num_classes: int, dtype: str = "float32") -> np.ndarray:
    """Converts a 1D array of indices to a one-hot encoded 2D array."""
    return np.eye(num_classes, dtype=dtype)[indices]


def shifted_softplus(
    x: np.ndarray, beta: float = 1.0, threshold: float = 20.0, shift: float = np.log(np.e - 1)
) -> np.ndarray:
    """Shifted version of the softplus function such that shifted_softplus(0) = 1"""
    return softplus(x + shift, beta=beta, threshold=threshold)


sigmoid = special.expit
softmax = special.softmax


def softplus(x: np.ndarray, beta: float = 1.0, threshold: float = 20.0) -> np.ndarray:
    """Numerically stabilized softplus function."""
    with np.errstate(over="ignore"):
        exp_beta_x = np.exp(beta * x)
    return np.where(beta * x > threshold, x, np.log1p(exp_beta_x) / beta)



from collections.abc import Sequence
from typing import TypeVar

import keras
import numpy as np

from bayesflow.types import Tensor
from . import logging

T = TypeVar("T")


def concatenate_valid(tensors: Sequence[Tensor | None], axis: int = 0) -> Tensor | None:
    """Concatenate multiple tensors along axis, ignoring None values."""
    tensors = [t for t in tensors if t is not None]

    if not tensors:
        return None

    return keras.ops.concatenate(tensors, axis=axis)


def expand(x: Tensor, n: int, side: str):
    if n < 0:
        raise ValueError(f"Cannot expand {n} times.")

    match side:
        case "left":
            idx = [None] * n + [...]
        case "right":
            idx = [...] + [None] * n
        case str() as name:
            raise ValueError(f"Invalid side {name!r}. Must be 'left' or 'right'.")
        case other:
            raise TypeError(f"Invalid side type {type(other)!r}. Must be str.")

    return x[tuple(idx)]


def expand_as(x: Tensor, y: Tensor, side: str):
    return expand_to(x, keras.ops.ndim(y), side)


def expand_to(x: Tensor, dim: int, side: str):
    return expand(x, dim - keras.ops.ndim(x), side)


def expand_left(x: Tensor, n: int) -> Tensor:
    """Expand x to the left n times"""
    if n < 0:
        raise ValueError(f"Cannot expand {n} times.")

    idx = [None] * n + [...]
    return x[tuple(idx)]


def expand_left_as(x: Tensor, y: Tensor) -> Tensor:
    """Expand x to the left, matching the dimension of y"""
    return expand_left_to(x, keras.ops.ndim(y))


def expand_left_to(x: Tensor, dim: int) -> Tensor:
    """Expand x to the left, matching dim"""
    return expand_left(x, dim - keras.ops.ndim(x))


def expand_right(x: Tensor, n: int) -> Tensor:
    """Expand x to the right n times"""
    if n < 0:
        raise ValueError(f"Cannot expand {n} times.")

    idx = [...] + [None] * n
    return x[tuple(idx)]


def expand_right_as(x: Tensor, y: Tensor) -> Tensor:
    """Expand x to the right, matching the dimension of y"""
    return expand_right_to(x, keras.ops.ndim(y))


def expand_right_to(x: Tensor, dim: int) -> Tensor:
    """Expand x to the right, matching dim"""
    return expand_right(x, dim - keras.ops.ndim(x))


def expand_tile(x: Tensor, n: int, axis: int) -> Tensor:
    """Expand and tile x along the given axis n times"""
    if keras.ops.is_tensor(x):
        x = keras.ops.expand_dims(x, axis)
    else:
        x = np.expand_dims(x, axis)

    return tile_axis(x, n, axis=axis)


def is_symbolic_tensor(x: Tensor) -> bool:
    if keras.utils.is_keras_tensor(x):
        return True

    match keras.backend.backend():
        case "jax":
            import jax

            return not jax.core.is_concrete(x)
        case "tensorflow":
            import tensorflow as tf

            return tf.is_symbolic_tensor(x)
        case "torch":
            return False
        case _:
            raise NotImplementedError(f"Symbolic tensor check not implemented for backend {keras.backend.backend()!r}")


def pad(x: Tensor, value: float | Tensor, n: int, axis: int, side: str = "both") -> Tensor:
    """
    Pad x with n values along axis on the given side.
    The pad value must broadcast against the shape of x, except for the pad axis, where it must broadcast against n.
    """
    if not keras.ops.is_tensor(value):
        value = keras.ops.full((), value, dtype=keras.ops.dtype(x))

    shape = list(keras.ops.shape(x))
    shape[axis] = n

    p = keras.ops.broadcast_to(value, shape)
    match side:
        case "left":
            return keras.ops.concatenate([p, x], axis=axis)
        case "right":
            return keras.ops.concatenate([x, p], axis=axis)
        case "both":
            return keras.ops.concatenate([p, x, p], axis=axis)
        case str() as name:
            raise ValueError(f"Invalid side {name!r}. Must be 'left', 'right', or 'both'.")
        case _:
            raise TypeError(f"Invalid side type {type(side)!r}. Must be str.")


def weighted_mean(elements: Tensor, weights: Tensor = None) -> Tensor:
    """
    Compute the (optionally) weighted mean of the input tensor.

    Parameters
    ----------
    elements : Tensor
        A tensor containing the elements to average.
    weights : Tensor, optional
        A tensor of the same shape as `elements` representing weights.
        If None, the mean is computed without weights.

    Returns
    -------
    Tensor
        A scalar tensor representing the (weighted) mean.
    """
    return keras.ops.mean(elements * weights if weights is not None else elements)


def searchsorted(sorted_sequence: Tensor, values: Tensor, side: str = "left") -> Tensor:
    """
    Find indices where elements should be inserted to maintain order.
    """

    match keras.backend.backend():
        case "jax":
            import jax
            import jax.numpy as jnp

            logging.warn_once(f"searchsorted is not yet optimized for backend {keras.backend.backend()!r}")

            # do not vmap over the side argument (we have to pass it as a positional argument)
            in_axes = [0, 0, None]

            # vmap over the batch dimension
            vss = jax.vmap(jnp.searchsorted, in_axes=in_axes)

            # flatten all batch dimensions
            ss = sorted_sequence.reshape((-1,) + sorted_sequence.shape[-1:])
            v = values.reshape((-1,) + values.shape[-1:])

            # noinspection PyTypeChecker
            indices = vss(ss, v, side)

            # restore the batch dimensions
            indices = indices.reshape(values.shape)

            # noinspection PyTypeChecker
            return indices
        case "tensorflow":
            import tensorflow as tf

            # always use int64 to avoid complicated graph code
            indices = tf.searchsorted(sorted_sequence, values, side=side, out_type="int64")

            return indices
        case "torch":
            import torch

            out_int32 = len(sorted_sequence) <= np.iinfo(np.int32).max

            indices = torch.searchsorted(
                sorted_sequence.contiguous(), values.contiguous(), side=side, out_int32=out_int32
            )

            return indices
        case _:
            raise NotImplementedError(f"Searchsorted not implemented for backend {keras.backend.backend()!r}")


def size_of(x) -> int:
    """
    :param x: A nested structure of tensors.
    :return: The total memory footprint of x, ignoring view semantics, in bytes.
    """
    if keras.ops.is_tensor(x) or isinstance(x, np.ndarray):
        return int(keras.ops.size(x)) * np.dtype(keras.ops.dtype(x)).itemsize

    # flatten nested structure
    x = keras.tree.flatten(x)

    # get unique tensors by id
    x = {id(tensor): tensor for tensor in x}

    # sum up individual sizes
    return sum(size_of(tensor) for tensor in x.values())


def stack_valid(tensors: Sequence[Tensor | None], axis: int = 0) -> Tensor | None:
    """Stack multiple tensors along axis, ignoring None values."""
    tensors = [t for t in tensors if t is not None]

    if not tensors:
        return None

    return keras.ops.stack(tensors, axis=axis)


def tile_axis(x: Tensor, n: int, axis: int) -> Tensor:
    """Tile x along the given axis n times"""
    repeats = [1] * keras.ops.ndim(x)
    repeats[axis] = n

    if keras.ops.is_tensor(x):
        return keras.ops.tile(x, repeats)

    return np.tile(x, repeats)


# we want to annotate this as Sequence[PyTree[Tensor]], but static type checkers do not support PyTree's type expansion
def tree_concatenate(structures: Sequence[T], axis: int = 0, numpy: bool = None) -> T:
    """Concatenate all tensors in the given sequence of nested structures.
    All objects in the given sequence must have the same structure.
    The output will adhere to this structure.

    :param structures: A sequence of nested structures of tensors.
        All structures in the sequence must have the same layout.
        Tensors in the same layout location must have compatible shapes for concatenation.
    :param axis: The axis along which to concatenate tensors.
    :param numpy: Whether to use numpy or keras for concatenation.
        Will convert all items in the structures to numpy arrays if True, tensors otherwise.
        Defaults to True if all tensors are numpy arrays, False otherwise.
    :return: A structure of concatenated tensors with the same layout as each input structure.
    """
    if numpy is None:
        numpy = not any(keras.tree.flatten(keras.tree.map_structure(keras.ops.is_tensor, structures)))

    if numpy:
        structures = keras.tree.map_structure(keras.ops.convert_to_numpy, structures)

        def concat(*items):
            return np.concatenate(items, axis=axis)
    else:
        structures = keras.tree.map_structure(keras.ops.convert_to_tensor, structures)

        def concat(*items):
            return keras.ops.concatenate(items, axis=axis)

    return keras.tree.map_structure(concat, *structures)


def tree_stack(structures: Sequence[T], axis: int = 0, numpy: bool = None) -> T:
    """Like :func:`tree_concatenate`, except tensors are stacked instead of concatenated."""
    if numpy is None:
        numpy = not any(keras.tree.flatten(keras.tree.map_structure(keras.ops.is_tensor, structures)))

    if numpy:
        structures = keras.tree.map_structure(keras.ops.convert_to_numpy, structures)

        def stack(*items):
            return np.stack(items, axis=axis)
    else:
        structures = keras.tree.map_structure(keras.ops.convert_to_tensor, structures)

        def stack(*items):
            return keras.ops.stack(items, axis=axis)

    return keras.tree.map_structure(stack, *structures)


def fill_triangular_matrix(x: Tensor, upper: bool = False, positive_diag: bool = False):
    """
    Reshapes a batch of matrix elements into a triangular matrix (either upper or lower).

    Note: If final axis has length 1, this simply reshapes to (batch_size, 1, 1) and optionally applies softplus.

    Parameters
    ----------
    x : Tensor of shape (batch_size, m)
        Batch of flattened nonzero matrix elements for triangular matrix.
    upper : bool
        Return upper triangular matrix if True, else lower triangular matrix. Default is False.
    positive_diag : bool
        Whether to apply a softplus operation to diagonal elements. Default is False.

    Returns
    -------
    Tensor of shape (batch_size, n, n)
        Batch of triangular matrices with m = n * (n + 1) / 2 unique nonzero elements.

    Raises
    ------
    ValueError
        If provided nonzero elements do not correspond to possible triangular matrix shape
        (n,n) with n = sqrt( 1/4 + 2 * m) - 1/2 due to m = n * (n + 1) / 2.
    """
    batch_shape = x.shape[:-1]
    m = x.shape[-1]

    if m == 1:
        y = keras.ops.reshape(x, (-1, 1, 1))
        if positive_diag:
            y = keras.activations.softplus(y)
        return y

    # Calculate matrix shape
    n = (0.25 + 2 * m) ** 0.5 - 0.5
    if not np.isclose(np.floor(n), n):
        raise ValueError(f"Input right-most shape ({m}) does not correspond to a triangular matrix.")
    else:
        n = int(n)

    # Trick: Create triangular matrix by concatenating with a flipped version of its tail, then reshape.
    x_tail = keras.ops.take(x, indices=list(range((m - (n**2 - m)), x.shape[-1])), axis=-1)
    if not upper:
        y = keras.ops.concatenate([x_tail, keras.ops.flip(x, axis=-1)], axis=len(batch_shape))
        y = keras.ops.reshape(y, (-1, n, n))
        y = keras.ops.tril(y)

        if positive_diag:
            y_offdiag = keras.ops.tril(y, k=-1)
            # carve out diagonal, by setting upper and lower offdiagonals to zero
            y_diag = keras.ops.tril(
                keras.ops.triu(keras.activations.softplus(y)),  # apply softplus to enforce positivity
            )
            y = y_diag + y_offdiag

    else:
        y = keras.ops.concatenate([x, keras.ops.flip(x_tail, axis=-1)], axis=len(batch_shape))
        y = keras.ops.reshape(y, (-1, n, n))
        y = keras.ops.triu(
            y,
        )

        if positive_diag:
            y_offdiag = keras.ops.triu(y, k=1)
            # carve out diagonal, by setting upper and lower offdiagonals to zero
            y_diag = keras.ops.tril(
                keras.ops.triu(keras.activations.softplus(y)),  # apply softplus to enforce positivity
            )
            y = y_diag + y_offdiag

    return y



import numpy as np
from scipy.special import expit

from .benchmark_simulator import BenchmarkSimulator


class BernoulliGLM(BenchmarkSimulator):
    def __init__(self, T: int = 100, scale_by_T: bool = True, rng: np.random.Generator = None):
        """Bernoulli GLM simulated benchmark
        See: https://arxiv.org/pdf/2101.04653.pdf, Task T.5

        Important: `scale_sum` should be set to False if the simulator is used
        with variable `T` during training, otherwise the information of `T` will
        be lost.

        Parameters
        ----------
        T: int, optional, default: 100
            The simulated duration of the task (eq. the number of Bernoulli draws).
        scale_by_T: bool, optional, default: True
            A flag indicating whether to scale the summayr statistics by T.
        rng: np.random.Generator or None, optional, default: None
            An optional random number generator to use.
        """

        self.T = T
        self.scale_by_T = scale_by_T
        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()

        # Covariance matrix computed once for efficiency
        F = np.zeros((9, 9))
        i = np.arange(9)
        F[i, i] = 1 + np.sqrt(i / 9)
        F[i[1:], i[:-1]] = -2
        F[i[2:], i[:-2]] = 1
        self.cov = np.linalg.inv(F.T @ F)

    def prior(self):
        """Generates a random draw from the custom prior over the 10
        Bernoulli GLM parameters (1 intercept and 9 weights). Uses a
        global covariance matrix `Cov` for the multivariate Gaussian prior
        over the model weights, which is pre-computed for efficiency.

        Returns
        -------
        params: np.ndarray of shape (10, )
            A single draw from the prior.
        """

        beta = self.rng.normal(0, 2)
        f = self.rng.multivariate_normal(np.zeros(9), self.cov)
        return np.append(beta, f)

    def observation_model(self, params: np.ndarray):
        """Simulates data from the custom Bernoulli GLM likelihood.

        Parameters
        ----------
        params: np.ndarray of shape (10, )
            The vector of model parameters (`params[0]` is intercept, `params[i], i > 0` are weights).

        Returns
        -------
        x: np.ndarray of shape (10, )
            The vector of sufficient summary statistics of the data.
        """

        # Unpack parameters
        beta, f = params[0], params[1:]

        # Generate design matrix
        V = self.rng.normal(size=(9, self.T))

        # Draw from Bernoulli GLM
        z = self.rng.binomial(n=1, p=expit(V.T @ f + beta))

        # Compute and return (scaled) sufficient summary statistics
        x1 = np.sum(z)
        x_rest = V @ z
        x = np.append(x1, x_rest)
        if self.scale_by_T:
            x /= self.T
        return x



import keras

from bayesflow.utils.serialization import deserialize, serialize

from ...utils import assert_models_equal


def test_serialize_deserialize(mlp, build_shapes):
    mlp.build(**build_shapes)

    serialized = serialize(mlp)
    deserialized = deserialize(serialized)
    reserialized = serialize(deserialized)

    assert reserialized == serialized


def test_save_and_load(tmp_path, mlp, build_shapes):
    mlp.build(**build_shapes)

    keras.saving.save_model(mlp, tmp_path / "model.keras")
    loaded = keras.saving.load_model(tmp_path / "model.keras")

    assert_models_equal(mlp, loaded)



import pytest
from tests.utils import check_combination_simulator_adapter


@pytest.fixture()
def batch_size():
    return 8


@pytest.fixture()
def num_samples():
    return 100


@pytest.fixture()
def summary_network():
    return None


@pytest.fixture()
def inference_network():
    from bayesflow.networks import CouplingFlow

    return CouplingFlow(subnet="mlp", depth=2, subnet_kwargs=dict(widths=(32, 32)))


@pytest.fixture()
def continuous_approximator(adapter, inference_network, summary_network):
    from bayesflow import ContinuousApproximator

    return ContinuousApproximator(
        adapter=adapter,
        inference_network=inference_network,
        summary_network=summary_network,
    )


@pytest.fixture()
def point_inference_network():
    from bayesflow.networks import PointInferenceNetwork
    from bayesflow.scores import NormedDifferenceScore, QuantileScore, MultivariateNormalScore

    return PointInferenceNetwork(
        scores=dict(
            mean=NormedDifferenceScore(k=2),
            quantiles=QuantileScore(q=[0.1, 0.5, 0.9]),
            mvn=MultivariateNormalScore(),
        ),
        subnet="mlp",
        subnet_kwargs=dict(widths=(32, 32)),
    )


@pytest.fixture()
def point_inference_network_with_multiple_parametric_scores():
    from bayesflow.networks import PointInferenceNetwork
    from bayesflow.scores import MultivariateNormalScore

    return PointInferenceNetwork(
        scores=dict(
            mvn1=MultivariateNormalScore(),
            mvn2=MultivariateNormalScore(),
        ),
    )


@pytest.fixture()
def point_approximator(adapter, point_inference_network, summary_network):
    from bayesflow import PointApproximator

    return PointApproximator(
        adapter=adapter,
        inference_network=point_inference_network,
        summary_network=summary_network,
    )


@pytest.fixture()
def point_approximator_with_multiple_parametric_scores(
    adapter, point_inference_network_with_multiple_parametric_scores, summary_network
):
    from bayesflow import PointApproximator

    return PointApproximator(
        adapter=adapter,
        inference_network=point_inference_network_with_multiple_parametric_scores,
        summary_network=summary_network,
    )


@pytest.fixture(
    params=["continuous_approximator", "point_approximator", "point_approximator_with_multiple_parametric_scores"],
    scope="function",
)
def approximator(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def adapter_without_sample_weight():
    from bayesflow import ContinuousApproximator

    return ContinuousApproximator.build_adapter(
        inference_variables=["mean", "std"],
        inference_conditions=["x"],
    )


@pytest.fixture()
def adapter_with_sample_weight():
    from bayesflow import ContinuousApproximator

    return ContinuousApproximator.build_adapter(
        inference_variables=["mean", "std"],
        inference_conditions=["x"],
        sample_weight="weight",
    )


@pytest.fixture(params=["adapter_without_sample_weight", "adapter_with_sample_weight"])
def adapter(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def normal_simulator():
    from tests.utils.normal_simulator import NormalSimulator

    return NormalSimulator()


@pytest.fixture()
def normal_simulator_with_sample_weight():
    from tests.utils.normal_simulator import NormalSimulator
    from bayesflow import make_simulator

    def weight(mean):
        return dict(weight=1.0)

    return make_simulator([NormalSimulator(), weight])


@pytest.fixture(params=["normal_simulator", "normal_simulator_with_sample_weight"])
def simulator(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def train_dataset(batch_size, adapter, simulator):
    check_combination_simulator_adapter(simulator, adapter)

    from bayesflow import OfflineDataset

    num_batches = 4
    data = simulator.sample((num_batches * batch_size,))
    return OfflineDataset(data=data, adapter=adapter, batch_size=batch_size, workers=4, max_queue_size=num_batches)


@pytest.fixture()
def validation_dataset(batch_size, adapter, simulator):
    from bayesflow import OfflineDataset

    num_batches = 2
    data = simulator.sample((num_batches * batch_size,))
    return OfflineDataset(data=data, adapter=adapter, batch_size=batch_size, workers=4, max_queue_size=num_batches)



from collections.abc import Sequence

import keras

from bayesflow.types import Tensor
from bayesflow.utils import sequential_kwargs
from bayesflow.utils.serialization import deserialize, serializable, serialize


@serializable
class Residual(keras.Sequential):
    def __init__(self, *layers: keras.Layer, **kwargs):
        if len(layers) == 1 and isinstance(layers[0], Sequence):
            layers = layers[0]
        super().__init__(list(layers), **sequential_kwargs(kwargs))
        self.projector = keras.layers.Dense(units=None, name="projector")

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self):
        base_config = super().get_config()
        base_config = sequential_kwargs(base_config)

        config = {
            "layers": self.layers,
            "projector": self.projector,
        }

        return base_config | serialize(config)

    def build(self, input_shape=None):
        if self.built:
            # building when the network is already built can cause issues with serialization
            # see https://github.com/keras-team/keras/issues/21147
            return

        # we have to avoid calling super().build() because this causes
        # shape errors when building on non-sets but doing inference on sets
        # this is a work-around for https://github.com/keras-team/keras/issues/21158
        output_shape = input_shape
        for layer in self._layers:
            layer.build(output_shape)
            output_shape = layer.compute_output_shape(output_shape)

        self.projector.units = output_shape[-1]
        self.projector.build(input_shape)

    def call(self, x: Tensor, training: bool = None, mask: bool = None) -> Tensor:
        return self.projector(x) + super().call(x, training=training, mask=mask)



import keras
from keras import ops

import numpy as np

import warnings

from bayesflow.types import Tensor
from bayesflow.utils import find_network, layer_kwargs, weighted_mean
from bayesflow.utils.serialization import deserialize, serializable, serialize

from ..inference_network import InferenceNetwork


@serializable
class ConsistencyModel(InferenceNetwork):
    """Implements a Consistency Model with Consistency Training (CT) a described in [1-2]. The adaptations to CT
    described in [2] were taken into account in our implementation for ABI [3].

    [1] Song, Y., Dhariwal, P., Chen, M. & Sutskever, I. (2023). Consistency Models. arXiv preprint arXiv:2303.01469

    [2] Song, Y., & Dhariwal, P. (2023). Improved Techniques for Training Consistency Models.
    arXiv preprint arXiv:2310.14189. Discussion: https://openreview.net/forum?id=WNzy9bRDvG

    [3] Schmitt, M., Pratz, V., Köthe, U., Bürkner, P. C., & Radev, S. T. (2023). Consistency models for scalable and
    fast simulation-based inference. arXiv preprint arXiv:2312.05440.
    """

    MLP_DEFAULT_CONFIG = {
        "widths": (256, 256, 256, 256, 256),
        "activation": "mish",
        "kernel_initializer": "he_normal",
        "residual": True,
        "dropout": 0.05,
        "spectral_normalization": False,
    }

    def __init__(
        self,
        total_steps: int | float,
        subnet: str | keras.Layer = "mlp",
        max_time: int | float = 200,
        sigma2: float = 1.0,
        eps: float = 0.001,
        s0: int | float = 10,
        s1: int | float = 50,
        subnet_kwargs: dict[str, any] = None,
        **kwargs,
    ):
        """Creates an instance of a consistency model (CM) to be used for standalone consistency training (CT).

        Parameters
        ----------
        total_steps : int
            The total number of training steps, must be calculated as number of epochs * number of batches
            and cannot be inferred during construction time.
        subnet      : str or type, optional, default: "mlp"
            A neural network type for the consistency model, will be
            instantiated using subnet_kwargs.
        max_time : int or float, optional, default: 200.0
            The maximum time of the diffusion
        sigma2      : float or Tensor of dimension (input_dim, 1), optional, default: 1.0
            Controls the shape of the skip-function
        eps         : float, optional, default: 0.001
            The minimum time
        s0          : int or float, optional, default: 10
            Initial number of discretization steps
        s1          : int or float, optional, default: 50
            Final number of discretization steps
        subnet_kwargs: dict[str, any], optional
            Keyword arguments passed to the subnet constructor or used to update the default MLP settings.
        **kwargs    : dict, optional, default: {}
            Additional keyword arguments
        """
        super().__init__(base_distribution="normal", **kwargs)

        self.total_steps = float(total_steps)

        if subnet_kwargs:
            warnings.warn(
                "Using `subnet_kwargs` is deprecated."
                "Instead, instantiate the network yourself and pass the arguments directly.",
                DeprecationWarning,
            )

        subnet_kwargs = subnet_kwargs or {}
        if subnet == "mlp":
            subnet_kwargs = ConsistencyModel.MLP_DEFAULT_CONFIG | subnet_kwargs

        self.subnet = find_network(subnet, **subnet_kwargs)
        self.output_projector = keras.layers.Dense(
            units=None, bias_initializer="zeros", kernel_initializer="zeros", name="output_projector"
        )

        self.sigma2 = ops.convert_to_tensor(sigma2)
        self.sigma = ops.sqrt(sigma2)
        self.eps = eps
        self.max_time = max_time
        self.c_huber = None
        self.c_huber2 = None

        self.s0 = float(s0)
        self.s1 = float(s1)

        # create variable that works with JIT compilation
        self.current_step = self.add_weight(name="current_step", initializer="zeros", trainable=False, dtype="int")
        self.current_step.assign(0)

        self.seed_generator = keras.random.SeedGenerator()

    @property
    def student(self):
        return self.subnet

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self):
        base_config = super().get_config()
        base_config = layer_kwargs(base_config)

        config = {
            "total_steps": self.total_steps,
            "subnet": self.subnet,
            "max_time": self.max_time,
            "sigma2": self.sigma2,
            "eps": self.eps,
            "s0": self.s0,
            "s1": self.s1,
            # we do not need to store subnet_kwargs
        }

        return base_config | serialize(config)

    def _schedule_discretization(self, step) -> float:
        """Schedule function for adjusting the discretization level `N` during
        the course of training.

        Implements the function N(k) from [2], Section 3.4.
        """

        k_ = ops.floor(self.total_steps / (ops.log(self.s1 / self.s0) / ops.log(2.0) + 1.0))
        out = ops.minimum(self.s0 * ops.power(2.0, ops.floor(step / k_)), self.s1) + 1.0
        return out

    def _discretize_time(self, num_steps, rho=7.0):
        """Function for obtaining the discretized time according to [2],
        Section 2, bottom of page 2.
        """

        N = num_steps + 1
        indices = ops.arange(1, N + 1, dtype="float32")
        one_over_rho = 1.0 / rho
        discretized_time = (
            self.eps**one_over_rho
            + (indices - 1.0) / (ops.cast(N, "float32") - 1.0) * (self.max_time**one_over_rho - self.eps**one_over_rho)
        ) ** rho
        return discretized_time

    def build(self, xz_shape, conditions_shape=None):
        if self.built:
            # building when the network is already built can cause issues with serialization
            # see https://github.com/keras-team/keras/issues/21147
            return

        self.base_distribution.build(xz_shape)

        self.output_projector.units = xz_shape[-1]

        input_shape = list(xz_shape)

        # time vector
        input_shape[-1] += 1

        if conditions_shape is not None:
            input_shape[-1] += conditions_shape[-1]

        input_shape = tuple(input_shape)

        self.subnet.build(input_shape)

        input_shape = self.subnet.compute_output_shape(input_shape)
        self.output_projector.build(input_shape)

        # Choose coefficient according to [2] Section 3.3
        self.c_huber = 0.00054 * ops.sqrt(xz_shape[-1])
        self.c_huber2 = self.c_huber**2

        # Calculate discretization schedule in advance
        # The Jax compiler requires fixed-size arrays, so we have
        # to store all the discretized_times in one matrix in advance
        # and later only access the relevant entries.

        # First, we calculate all unique numbers of discretization steps n
        # in a loop, as self.total_steps might be large
        self.max_n = int(self._schedule_discretization(self.total_steps))

        if self.max_n != self.s1 + 1:
            raise ValueError("The maximum number of discretization steps must be equal to s1 + 1.")

        unique_n = set()
        for step in range(int(self.total_steps)):
            unique_n.add(int(self._schedule_discretization(step)))
        unique_n = sorted(list(unique_n))

        # Next, we calculate the discretized times for each n
        # and establish a mapping between n and the position i of the
        # discretizated times in the vector
        discretized_times = np.zeros((len(unique_n), self.max_n + 1))
        discretization_map = np.zeros((self.max_n + 1,), dtype=np.int32)
        for i, n in enumerate(unique_n):
            disc = ops.convert_to_numpy(self._discretize_time(n))
            discretized_times[i, : len(disc)] = disc
            discretization_map[n] = i

        # Finally, we convert the vectors to tensors
        self.discretized_times = ops.convert_to_tensor(discretized_times, dtype="float32")
        self.discretization_map = ops.convert_to_tensor(discretization_map)

    def _forward_train(
        self, x: Tensor, noise: Tensor, t: Tensor, conditions: Tensor = None, training: bool = False, **kwargs
    ) -> Tensor:
        """Forward function for training. Calls consistency function with noisy input"""
        inp = x + t * noise
        return self.consistency_function(inp, t, conditions=conditions, training=training)

    def _forward(self, x: Tensor, conditions: Tensor = None, **kwargs) -> Tensor:
        # Consistency Models only learn the direction from noise distribution
        # to target distribution, so we cannot implement this function.
        raise NotImplementedError("Consistency Models are not invertible")

    def _inverse(self, z: Tensor, conditions: Tensor = None, training: bool = False, **kwargs) -> Tensor:
        """Generate random draws from the approximate target distribution
        using the multistep sampling algorithm from [1], Algorithm 1.

        Parameters
        ----------
        z           : Tensor
            Samples from a standard normal distribution
        conditions  : Tensor, optional, default: None
            Conditions for the approximate conditional distribution
        training    : bool, optional, default: True
            Whether internal layers (e.g., dropout) should behave in train or inference mode.
        **kwargs    : dict, optional, default: {}
            Additional keyword arguments. Include `steps` (default: 10) to
            adjust the number of sampling steps.

        Returns
        -------
        x            : Tensor
            The approximate samples
        """
        steps = kwargs.get("steps", 10)
        x = keras.ops.copy(z) * self.max_time
        discretized_time = keras.ops.flip(self._discretize_time(steps), axis=-1)
        t = keras.ops.full((*keras.ops.shape(x)[:-1], 1), discretized_time[0], dtype=x.dtype)

        x = self.consistency_function(x, t, conditions=conditions, training=training)

        for n in range(1, steps):
            noise = keras.random.normal(keras.ops.shape(x), dtype=keras.ops.dtype(x), seed=self.seed_generator)
            x_n = x + keras.ops.sqrt(keras.ops.square(discretized_time[n]) - self.eps**2) * noise
            t = keras.ops.full_like(t, discretized_time[n])
            x = self.consistency_function(x_n, t, conditions=conditions, training=training)
        return x

    def consistency_function(self, x: Tensor, t: Tensor, conditions: Tensor = None, training: bool = False) -> Tensor:
        """Compute consistency function.

        Parameters
        ----------
        x           : Tensor
            Input vector
        t           : Tensor
            Vector of time samples in [eps, T]
        conditions  : Tensor
            The conditioning vector
        training    : bool, optional, default: True
            Whether internal layers (e.g., dropout) should behave in train or inference mode.
        """

        if conditions is not None:
            xtc = ops.concatenate([x, t, conditions], axis=-1)
        else:
            xtc = ops.concatenate([x, t], axis=-1)

        f = self.output_projector(self.subnet(xtc, training=training))

        # Compute skip and out parts (vectorized, since self.sigma2 is of shape (1, input_dim)
        # Thus, we can do a cross product with the time vector which is (batch_size, 1) for
        # a resulting shape of cskip and cout of (batch_size, input_dim)
        skip = self.sigma2 / ((t - self.eps) ** 2 + self.sigma2)
        out = self.sigma * (t - self.eps) / (ops.sqrt(self.sigma2 + t**2))

        out = skip * x + out * f
        return out

    def compute_metrics(
        self, x: Tensor, conditions: Tensor = None, sample_weight: Tensor = None, stage: str = "training"
    ) -> dict[str, Tensor]:
        base_metrics = super().compute_metrics(x, conditions=conditions, stage=stage)

        # The discretization schedule requires the number of passed training steps.
        # To be independent of external information, we track it here.
        if stage == "training":
            self.current_step.assign_add(1)
            self.current_step.assign(ops.minimum(self.current_step, self.total_steps - 1))

        discretization_index = ops.take(
            self.discretization_map, ops.cast(self._schedule_discretization(self.current_step), "int")
        )
        discretized_time = ops.take(self.discretized_times, discretization_index, axis=0)

        # Randomly sample t_n and t_[n+1] and reshape to (batch_size, 1)
        # adapted noise schedule from [2], Section 3.5
        p_mean = -1.1
        p_std = 2.0
        p = ops.where(
            discretized_time[1:] > 0.0,
            ops.erf((ops.log(discretized_time[1:]) - p_mean) / (ops.sqrt(2.0) * p_std))
            - ops.erf((ops.log(discretized_time[:-1]) - p_mean) / (ops.sqrt(2.0) * p_std)),
            0.0,
        )

        log_p = ops.log(p)
        times = keras.random.categorical(ops.expand_dims(log_p, 0), ops.shape(x)[0], seed=self.seed_generator)[0]
        t1 = ops.take(discretized_time, times)[..., None]
        t2 = ops.take(discretized_time, times + 1)[..., None]

        # generate noise vector
        noise = keras.random.normal(keras.ops.shape(x), dtype=keras.ops.dtype(x), seed=self.seed_generator)

        teacher_out = self._forward_train(x, noise, t1, conditions=conditions, training=stage == "training")
        # difference between teacher and student: different time,
        # and no gradient for the teacher
        teacher_out = ops.stop_gradient(teacher_out)
        student_out = self._forward_train(x, noise, t2, conditions=conditions, training=stage == "training")

        # weighting function, see [2], Section 3.1
        lam = 1 / (t2 - t1)

        # Pseudo-huber loss, see [2], Section 3.3
        loss = lam * (ops.sqrt(ops.square(teacher_out - student_out) + self.c_huber2) - self.c_huber)
        loss = weighted_mean(loss, sample_weight)

        return base_metrics | {"loss": loss}



import numpy as np

from bayesflow.utils.serialization import serializable

from .elementwise_transform import ElementwiseTransform


@serializable
class AsSet(ElementwiseTransform):
    """The `.as_set(["x", "y"])` transform indicates that both `x` and `y` are treated as sets.

    That is, their values will be treated as *exchangable* such that they will imply
    the same inference regardless of the values' order.
    This is useful, for example, in a linear regression context where we can index
    the observations in arbitrary order and always get the same regression line.

    Currently, all this transform does is to ensure that the variable
    arrays are at least 3D. The 2rd dimension is treated as the
    set dimension and the 3rd dimension as the data dimension.
    In the future, the transform will have more advanced behavior
    to better ensure the correct treatment of sets.

    Examples
    --------
    >>> adapter = bf.Adapter().as_set(["x", "y"])
    """

    def forward(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return np.atleast_3d(data)

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        if data.shape[2] == 1:
            return np.squeeze(data, axis=2)

        return data

    def get_config(self) -> dict:
        return {}



from .maximum_mean_discrepancy import maximum_mean_discrepancy
from .root_mean_squared_error import root_mean_squared_error

from ...utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])



import numpy as np
from scipy.integrate import odeint

from .benchmark_simulator import BenchmarkSimulator


class SIR(BenchmarkSimulator):
    def __init__(
        self,
        N: float = 1e6,
        T: int = 160,
        I0: float = 1.0,
        R0: float = 0.0,
        subsample: int = None,
        total_count: int = 1000,
        scale_by_total: bool = True,
        rng: np.random.Generator = None,
    ):
        """SIR simulated benchmark
        See: https://arxiv.org/pdf/2101.04653.pdf, Task T.9

        NOTE: the simulator scales outputs between 0 and 1.

        Parameters
        ----------
        N: float, optional, default: 1e6
            The size of the simulated population.
        T: int, optional, default: 160
            The duration (time horizon) of the simulation.
        I0: float, optional, default: 1.0
            The number of initially infected individuals.
        R0: float, optional, default: 0.0
            The number of initially recovered individuals.
        subsample: int or None, optional, default: 10
            The number of evenly spaced time points to return. If `None`,
            no subsampling will be performed and all `T` timepoints will be returned.
        total_count: int, optional, default: 1000
            The `N` parameter of the binomial noise distribution. Used just
            for scaling the data and magnifying the effect of noise, such that
            `max infected == total_count`.
        scale_by_total: bool, optional, default: True
            Scales the outputs by `total_count` if set to True.
        rng: np.random.Generator or None, optional, default: None
            An optional random number generator to use.
        """

        self.N = N
        self.T = T
        self.I0 = I0
        self.R0 = R0
        self.subsample = subsample
        self.total_count = total_count
        self.scale_by_total = scale_by_total
        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()

    def _deriv(self, x, t, N, beta, gamma):
        """Helper function for scipy.integrate.odeint."""

        s, i, r = x
        dS = -beta * s * i / N
        dI = beta * s * i / N - gamma * i
        dR = gamma * i
        return dS, dI, dR

    def prior(self):
        """Generates a random draw from a 2-dimensional (independent) lognormal prior
        which represents the contact and recovery rate parameters of a basic SIR model.

        Returns
        -------
        params : np.ndarray of shape (2, )
            A single draw from the 2-dimensional prior.
        """

        return self.rng.lognormal(mean=[np.log(0.4), np.log(1 / 8)], sigma=[0.5, 0.2])

    def observation_model(self, params: np.ndarray):
        """Runs a basic SIR model simulation for T time steps and returns `subsample` evenly spaced
        points from the simulated trajectory, given disease parameters (contact and recovery rate) `params`.

        Parameters
        ----------
        params         : np.ndarray of shape (2,)
            The 2-dimensional vector of disease parameters.

        Returns
        -------
        x : np.ndarray of shape (subsample, ) or (T, 1) if subsample=None
            The time series of simulated infected individuals. A trailing dimension of 1 should
            be added by a BayesFlow configurator if the data is (properly) to be treated as time series.
        """

        # Create vector (list) of initial conditions
        x0 = self.N - self.I0 - self.R0, self.I0, self.R0

        # Unpack parameter vector into scalars
        beta, gamma = params

        # Prepate time vector between 0 and T of length T
        t_vec = np.linspace(0, self.T, self.T)

        # Integrate using scipy and retain only infected (2-nd dimension)
        irt = odeint(self._deriv, x0, t_vec, args=(self.N, beta, gamma))[:, 1]

        # Subsample evenly the specified number of points, if specified
        if self.subsample is not None:
            irt = irt[:: (self.T // self.subsample)]
        else:
            irt = irt[:, None]

        # Truncate irt, so that small underflow below zero becomes zero
        irt = np.maximum(irt, 0.0)

        # Add noise and scale, if indicated
        x = self.rng.binomial(n=self.total_count, p=irt / self.N)
        if self.scale_by_total:
            x = x / self.total_count

        return x



from collections.abc import Callable, Sequence, Mapping
from typing import Protocol

import numpy as np

from bayesflow.utils.serialization import serializable, serialize

from .elementwise_transform import ElementwiseTransform
from .transform import Transform


class Predicate(Protocol):
    def __call__(self, key: str, value: np.ndarray, inverse: bool) -> bool:
        raise NotImplementedError


@serializable
class FilterTransform(Transform):
    """
    Implements a transform that applies a different transform on a subset of the data.

    Used by other transforms and base adapter class.
    """

    def __init__(
        self,
        include: str | Sequence[str] = None,
        *,
        transform_constructor: Callable[..., ElementwiseTransform],
        predicate: Predicate = None,
        exclude: str | Sequence[str] = None,
        transform_map: Mapping[str, ElementwiseTransform] = None,
        **kwargs,
    ):
        super().__init__()

        if isinstance(include, str):
            include = [include]

        if isinstance(exclude, str):
            exclude = [exclude]

        self.transform_constructor = transform_constructor

        self.predicate = predicate
        self.include = include
        self.exclude = exclude

        self.kwargs = kwargs

        self.transform_map = transform_map or {}

    def __repr__(self):
        if e := self.extra_repr():
            return f"{self.transform_constructor.__name__}({e})"

        return self.transform_constructor.__name__

    def extra_repr(self) -> str:
        result = ""

        if self.predicate is not None:
            result += f"predicate={self.predicate.__name__}"

        if self.include is not None:
            if result:
                result += ", "

            result += f"include={self.include!r}"

        if self.exclude is not None:
            if result:
                result += ", "

            result += f"exclude={self.exclude!r}"

        return result

    def get_config(self) -> dict:
        config = {
            "include": self.include,
            "transform_constructor": self.transform_constructor,
            "predicate": self.predicate,
            "exclude": self.exclude,
            "transform_map": self.transform_map,
            **self.kwargs,
        }
        return serialize(config)

    def forward(self, data: dict[str, np.ndarray], *, strict: bool = True, **kwargs) -> dict[str, np.ndarray]:
        data = data.copy()

        if strict and self.include is not None:
            missing_keys = set(self.include) - set(data.keys())
            if missing_keys:
                raise KeyError(f"Missing keys from include list: {missing_keys!r}")

        for key, value in data.items():
            if self._should_transform(key, value, inverse=False):
                data[key] = self._apply_transform(key, value, inverse=False, **kwargs)

        return data

    def inverse(self, data: dict[str, np.ndarray], *, strict: bool = False, **kwargs) -> dict[str, np.ndarray]:
        data = data.copy()

        if strict and self.include is not None:
            missing_keys = set(self.include) - set(data.keys())
            if missing_keys:
                raise KeyError(f"Missing keys from include list: {missing_keys!r}")

        for key, value in data.items():
            if self._should_transform(key, value, inverse=True):
                data[key] = self._apply_transform(key, value, inverse=True)

        return data

    def _should_transform(self, key: str, value: np.ndarray, inverse: bool = False) -> bool:
        match self.predicate, self.include, self.exclude:
            case None, None, None:
                return True

            case None, None, exclude:
                return key not in exclude

            case None, include, None:
                return key in include

            case None, include, exclude:
                return key in include and key not in exclude

            case predicate, None, None:
                return predicate(key, value, inverse=inverse)

            case predicate, None, exclude:
                if key in exclude:
                    return False
                return predicate(key, value, inverse=inverse)

            case predicate, include, None:
                if key in include:
                    return True
                return predicate(key, value, inverse=inverse)

            case predicate, include, exclude:
                if key in exclude:
                    return False
                if key in include:
                    return True
                return predicate(key, value, inverse=inverse)

    def _apply_transform(self, key: str, value: np.ndarray, inverse: bool = False, **kwargs) -> np.ndarray:
        if key not in self.transform_map:
            self.transform_map[key] = self.transform_constructor(**self.kwargs)

        transform = self.transform_map[key]

        return transform(value, inverse=inverse, **kwargs)



import keras
import numpy as np
import pytest


@pytest.fixture()
def batch_size():
    return 16


@pytest.fixture()
def num_batches():
    return 4


@pytest.fixture(params=["online_dataset", "offline_dataset"])
def dataset(request, online_dataset, offline_dataset):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def model():
    class Model(keras.Model):
        def call(self, *args, **kwargs):
            pass

        def compute_loss(self, *args, **kwargs):
            return keras.ops.zeros(())

    model = Model()
    model.compile()

    return model


@pytest.fixture()
def offline_dataset(simulator, batch_size, num_batches, workers, use_multiprocessing):
    from bayesflow import OfflineDataset

    # TODO: there is a bug in keras where if len(dataset) == 1 batch
    #  fit will error because no logs are generated
    #  the single batch is then skipped entirely
    data = simulator.sample((batch_size * num_batches,))
    return OfflineDataset(
        data, batch_size=batch_size, workers=workers, use_multiprocessing=use_multiprocessing, adapter=None
    )


@pytest.fixture()
def online_dataset(simulator, batch_size, num_batches, workers, use_multiprocessing):
    from bayesflow import OnlineDataset

    return OnlineDataset(
        simulator,
        batch_size=batch_size,
        num_batches=num_batches,
        workers=workers,
        use_multiprocessing=use_multiprocessing,
        adapter=None,
    )


# these need to be global for pickle


class Simulator:
    def sample(self, batch_shape):
        return dict(x=np.random.standard_normal(size=batch_shape + (2,)).astype("float32"))


def sample_contexts_unbatched(**kwargs):
    return dict(r=np.random.standard_normal(), alpha=np.random.standard_normal())


def sample_parameters_unbatched(**kwargs):
    return dict(theta=np.random.standard_normal(size=2))


def sample_observables_unbatched(r, alpha, theta, **kwargs):
    return dict(x=np.random.standard_normal(size=2))


def sample_contexts_batched(shape, **kwargs):
    return dict(r=np.random.standard_normal(size=shape), alpha=np.random.standard_normal(size=shape))


def sample_parameters_batched(shape, **kwargs):
    return dict(theta=np.random.standard_normal(size=shape + (2,)))


def sample_observables_batched(shape, r, alpha, theta, **kwargs):
    return dict(x=np.random.standard_normal(size=shape + (2,)))


@pytest.fixture(params=["class", "unbatched_composite"])
def simulator(request):
    from bayesflow.simulators import make_simulator

    if request.param == "class":
        simulator = Simulator()
    elif request.param == "unbatched_composite":
        simulator = make_simulator(
            [sample_contexts_unbatched, sample_parameters_unbatched, sample_observables_unbatched]
        )
    else:
        raise NotImplementedError

    return simulator


@pytest.fixture(params=[False])
def use_multiprocessing(request):
    return request.param


@pytest.fixture(params=[1, 2])
def workers(request):
    return request.param



from collections.abc import Callable, Mapping, Sequence
import keras
import numpy as np

from bayesflow.types import Shape


def batched_call(
    f: callable,
    batch_shape: Shape,
    args: Sequence[any] = (),
    kwargs: Mapping[str, any] = None,
    map_predicate: Callable[[any], bool] = None,
    flatten: bool = False,
) -> list:
    """Map f over the given batch shape with a for loop, preserving randomness unlike the keras built-in map apis.

    :param f: The function to call.

    :param batch_shape: The shape of the batch.

    :param args: Any number and type of positional arguments to f.
        Arguments indicated by `map_predicate` will be indexed over the first len(batch_shape) axes.

    :param kwargs: Any number and type of keyword arguments to f.
        Arguments indicated by `map_predicate` will be indexed over the first len(batch_shape) axes.

    :param map_predicate: A function that returns True if an argument should be indexed over the batch shape.
        By default, all array-like arguments are mapped.

    :param flatten: Whether to flatten the output.

    :return: A list of outputs of f for each element in the batch.
    """
    if kwargs is None:
        kwargs = {}

    if map_predicate is None:

        def map_predicate(arg):
            if isinstance(arg, np.ndarray):
                return arg.ndim >= len(batch_shape)
            if keras.ops.is_tensor(arg):
                return keras.ops.ndim(arg) >= len(batch_shape)

            return False

    outputs = np.empty(batch_shape, dtype="object")

    for index in np.ndindex(batch_shape):
        map_args = [arg[index] if map_predicate(arg) else arg for arg in args]
        map_kwargs = {key: value[index] if map_predicate(value) else value for key, value in kwargs.items()}

        outputs[index] = f(*map_args, **map_kwargs)

    if flatten:
        outputs = outputs.flatten()

    return outputs.tolist()



import pytest
import keras

from bayesflow.networks.embeddings import (
    FourierEmbedding,
    RecurrentEmbedding,
    Time2Vec,
)


def test_fourier_embedding_output_shape_and_type():
    embed_dim = 8
    batch_size = 4

    emb_layer = FourierEmbedding(embed_dim=embed_dim, include_identity=True)
    # use keras.ops.zeros with shape (batch_size, 1) and float32 dtype
    t = keras.ops.zeros((batch_size, 1), dtype="float32")

    emb = emb_layer(t)
    # Expected shape is (batch_size, embed_dim + 1) if include_identity else (batch_size, embed_dim)
    expected_dim = embed_dim + 1
    assert emb.shape[0] == batch_size
    assert emb.shape[1] == expected_dim
    # Check type - it should be a Keras tensor, convert to numpy for checking
    np_emb = keras.ops.convert_to_numpy(emb)
    assert np_emb.shape == (batch_size, expected_dim)


def test_fourier_embedding_without_identity():
    embed_dim = 8
    batch_size = 3

    emb_layer = FourierEmbedding(embed_dim=embed_dim, include_identity=False)
    t = keras.ops.zeros((batch_size, 1), dtype="float32")

    emb = emb_layer(t)
    expected_dim = embed_dim
    assert emb.shape[0] == batch_size
    assert emb.shape[1] == expected_dim


def test_fourier_embedding_raises_for_odd_embed_dim():
    with pytest.raises(ValueError):
        FourierEmbedding(embed_dim=7)


def test_recurrent_embedding_lstm_and_gru_shapes():
    batch_size = 2
    seq_len = 5
    dim = 3
    embed_dim = 6

    # Dummy input
    x = keras.ops.zeros((batch_size, seq_len, dim), dtype="float32")

    # lstm
    lstm_layer = RecurrentEmbedding(embed_dim=embed_dim, embedding="lstm")
    emb_lstm = lstm_layer(x)
    # Check the concatenated shape: last dimension = original dim + embed_dim
    assert emb_lstm.shape == (batch_size, seq_len, dim + embed_dim)

    # gru
    gru_layer = RecurrentEmbedding(embed_dim=embed_dim, embedding="gru")
    emb_gru = gru_layer(x)
    assert emb_gru.shape == (batch_size, seq_len, dim + embed_dim)


def test_recurrent_embedding_raises_unknown_embedding():
    with pytest.raises(ValueError):
        RecurrentEmbedding(embed_dim=4, embedding="unknown")


def test_time2vec_shapes_and_output():
    batch_size = 3
    seq_len = 7
    dim = 2
    num_periodic_features = 4

    x = keras.ops.zeros((batch_size, seq_len, dim), dtype="float32")
    time2vec_layer = Time2Vec(num_periodic_features=num_periodic_features)

    emb = time2vec_layer(x)
    # The last dimension should be dim + num_periodic_features + 1 (trend + periodic)
    expected_dim = dim + num_periodic_features + 1
    assert emb.shape == (batch_size, seq_len, expected_dim)



import keras

from bayesflow.types import Tensor
from bayesflow.utils.decorators import sanitize_input_shape
from bayesflow.utils.serialization import serializable

from .mab import MultiHeadAttentionBlock


@serializable
class SetAttentionBlock(MultiHeadAttentionBlock):
    """Implements the SAB block from [1] which represents learnable self-attention.

    [1] Lee, J., Lee, Y., Kim, J., Kosiorek, A., Choi, S., & Teh, Y. W. (2019).
        Set transformer: A framework for attention-based permutation-invariant neural networks.
        In International conference on machine learning (pp. 3744-3753). PMLR.
    """

    # noinspection PyMethodOverriding
    @sanitize_input_shape
    def build(self, input_set_shape):
        self.call(keras.ops.zeros(input_set_shape))

    def call(self, input_set: Tensor, training: bool = False, **kwargs) -> Tensor:
        """Performs the forward pass through the self-attention layer.

        Parameters
        ----------
        input_set  : Tensor (e.g., np.ndarray, tf.Tensor, ...)
            Input of shape (batch_size, set_size, input_dim)
        training   : boolean, optional (default - True)
            Passed to the optional internal dropout and spectral normalization
            layers to distinguish between train and test time behavior.
        **kwargs   : dict, optional (default - {})
            Additional keyword arguments passed to the internal attention layer,
            such as ``attention_mask`` or ``return_attention_scores``

        Returns
        -------
        out : Tensor
            Output of shape (batch_size, set_size, output_dim)
        """

        return super().call(input_set, input_set, training=training, **kwargs)

    # noinspection PyMethodOverriding
    @sanitize_input_shape
    def compute_output_shape(self, input_set_shape):
        return keras.ops.shape(self.call(keras.ops.zeros(input_set_shape)))



from bayesflow.types import Tensor

from .exceptions import ShapeError


def check_lengths_same(*args):
    if len(set(map(len, args))) > 1:
        raise ValueError(f"All tuple arguments must have the same length, but lengths are {tuple(map(len, args))}.")


def check_prior_shapes(variables: Tensor):
    """
    Checks the shape of posterior draws as required by most diagnostic functions

    Parameters
    ----------
    variables     : Tensor of shape (num_data_sets, num_params)
        The prior_samples from generating num_data_sets
    """

    if len(variables.shape) != 2:
        raise ShapeError(
            "prior_samples samples should be a 2-dimensional array, with the "
            "first dimension being the number of (simulated) data sets / prior_samples draws "
            "and the second dimension being the number of variables, "
            f"but your input has dimensions {len(variables.shape)}"
        )


def check_estimates_shapes(variables: Tensor):
    """
    Checks the shape of model-generated predictions (posterior draws, point estimates)
    as required by most diagnostic functions

    Parameters
    ----------
    variables     : Tensor of shape (num_data_sets, num_post_draws, num_params)
        The prior_samples from generating num_data_sets
    """
    if len(variables.shape) != 2 and len(variables.shape) != 3:
        raise ShapeError(
            "estimates should be a 2- or 3-dimensional array, with the "
            "first dimension being the number of data sets, "
            "(optional) second dimension the number of posterior draws per data set, "
            "and the last dimension the number of estimated variables, "
            f"but your input has dimensions {len(variables.shape)}"
        )


def check_consistent_shapes(estimates: Tensor, prior_samples: Tensor):
    """
    Checks whether the model-generated predictions (posterior draws, point estimates) and
    prior_samples have consistent leading (num_data_sets) and trailing (num_params) dimensions
    """
    if estimates.shape[0] != prior_samples.shape[0]:
        raise ShapeError(
            "The number of elements over the first dimension of estimates and prior_samples"
            f"should match, but estimates have {estimates.shape[0]} and prior_samples has "
            f"{prior_samples.shape[0]} elements, respectively."
        )
    if estimates.shape[-1] != prior_samples.shape[-1]:
        raise ShapeError(
            "The number of elements over the last dimension of estimates and prior_samples"
            f"should match, but estimates has {estimates.shape[0]} and prior_samples has "
            f"{prior_samples.shape[0]} elements, respectively."
        )


def check_estimates_prior_shapes(estimates: Tensor, prior_samples: Tensor):
    """
    Checks requirements for the shapes of estimates and prior_samples draws as
    necessitated by most diagnostic functions.

    Parameters
    ----------
    estimates      : Tensor of shape (num_data_sets, num_post_draws, num_params) or (num_data_sets, num_params)
        The model-generated predictions (posterior draws, point estimates) obtained from num_data_sets
    prior_samples     : Tensor of shape (num_data_sets, num_params)
        The prior_samples draws obtained for generating num_data_sets

    Raises
    ------
    ShapeError
        If there is a deviation form the expected shapes of `estimates` and `estimates`.
    """

    check_estimates_shapes(estimates)
    check_prior_shapes(prior_samples)
    check_consistent_shapes(estimates, prior_samples)



import numpy as np

from bayesflow.utils.serialization import serializable, serialize

from .elementwise_transform import ElementwiseTransform


@serializable
class NumpyTransform(ElementwiseTransform):
    """
    A class to apply element-wise transformations using plain NumPy functions.

    Attributes
    ----------
    _forward : str
        The name of the NumPy function to apply in the forward transformation.
    _inverse : str
        The name of the NumPy function to apply in the inverse transformation.
    """

    #: Dict of `np.ufunc` that support automatic selection of their inverse.
    INVERSE_METHODS = {
        np.arctan: np.tan,
        np.exp: np.log,
        np.expm1: np.log1p,
        np.square: np.sqrt,
        np.reciprocal: np.reciprocal,
    }
    # ensure the map is symmetric
    INVERSE_METHODS |= {v: k for k, v in INVERSE_METHODS.items()}

    def __init__(self, forward: str, inverse: str = None):
        """
        Initializes the NumpyTransform with specified forward and inverse functions.

        Parameters
        ----------
        forward : str
            The name of the NumPy function to use for the forward transformation.
        inverse : str, optional
            The name of the NumPy function to use for the inverse transformation.
            By default, the inverse is inferred from the forward argument for supported methods.
        """
        super().__init__()

        if isinstance(forward, str):
            forward = getattr(np, forward)

        if not isinstance(forward, np.ufunc):
            raise ValueError("Forward transformation must be a NumPy Universal Function (ufunc).")

        if inverse is None:
            if forward not in self.INVERSE_METHODS:
                raise ValueError(f"Cannot infer inverse for method {forward!r}")

            inverse = self.INVERSE_METHODS[forward]

        if isinstance(inverse, str):
            inverse = getattr(np, inverse)

        if not isinstance(inverse, np.ufunc):
            raise ValueError("Inverse transformation must be a NumPy Universal Function (ufunc).")

        self._forward = forward
        self._inverse = inverse

    def get_config(self) -> dict:
        return serialize({"forward": self._forward.__name__, "inverse": self._inverse.__name__})

    def forward(self, data: dict[str, any], **kwargs) -> dict[str, any]:
        return self._forward(data)

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return self._inverse(data)
